/* main.cpp - Main file for a simple SIP UA.
 * (c) 1999-2003 Columbia University. All rights reserved.
*/
extern "C" {
#include <pthread.h>
#include "timespec.h"

#include "strdupn.h"

extern char *optarg;
}

#include <stdio.h>
#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <list>

#include "radio.h"  /* Audio tool API */

#include "iptelcall.h"
#include "sipmessage.h"
#include "iptelsession.h"
#include "sipcallobj.h"
#include "sipregobj.h"
#include "sipendpoint.h"
#include "siplibrary.h"
#include "sipsession.h"

#include "command.h"

#ifdef USE_GSOAP
#include "floorcontrol.h"
#include "soapFloorControlProxy.h"  //get proxy
#endif

using namespace std;

/* Include the read line libraroes.
 * If you have compiled it use --with-readline=/your/readline/dir 
 * configure option.
 * If you are using standard installation use --with-readline=default
 * The includes are different for standard and compiled
 * installation.
 */
#ifdef USE_READLINE
#include "readline/readline.h"
#else
#ifdef USE_COMPILED_READLINE
#define READLINE_LIBRARY 1
extern "C" {
#include "readline/readline.h"
}
#define USE_READLINE 1
#endif
#endif

//////////////////////////////// MACRO definitions ///////////////////////

#define SIPUA_HISTORY_FILE "sipua_history.txt"
#define SIPUA_HISTORY_COUNT 200

//////////////////////////////// Local class definitions /////////////////

/**
 * MySIPUserAgent is defined as a user agent or endpoint to handling
 * this SIP user agent.
 */
class MySIPUserAgent : public SIPSessionEndpoint
{
public:
  MySIPUserAgent();
  ~MySIPUserAgent();

  IptelCall *AllocateNewCallObject(const std::string callid);

  void Lock()    {pthread_mutex_lock(&new_call_mutex);};
  void Unlock()  {pthread_mutex_unlock(&new_call_mutex);};
  void Signal()  {pthread_cond_broadcast(&new_call_cond);};
  void Wait()    {pthread_cond_wait(&new_call_cond, &new_call_mutex);};
  void Wait(long time);

  friend class MySIPCall;

protected:
  /* cond and mutex for new incoming call signaling */
  pthread_cond_t new_call_cond;
  pthread_mutex_t new_call_mutex;

  /** Current call id counter. */
  int  call_num;
};

/**
 * MySIPCall defines call back functions to handle this
 * user agents incoming calls, termination, rejection etc.
 */
class MySIPCall : public SIPSession
{
public:
  MySIPCall(MySIPUserAgent & ep, const std::string callid = "");

  ~MySIPCall();

  void   OnNewIncomingCall(IptelEndpoint & ep);
  bool   OnHangup();
  void   OnCallEstablished();
  bool   OnCallRejected();
  bool   OnRemoteSessionChanged(const SessionDescription & sdp);
  bool   OnFailureResponse(SIPResponseMessage & msg);


  bool   OnSubscribe(const  SIPRequestMessage& request,
                     SIPResponseMessage& response);
  bool   OnSubscribeSuccess(const SIPResponseMessage& response);
  bool   OnSubscribeFailed(const SIPResponseMessage& response);
  bool   OnSubscribeExpired();
  bool   OnNotifySuccess(const SIPResponseMessage& response);
  bool   OnNotifyFailed(const SIPResponseMessage& response);
  void   OnNotify(const SIPRequestMessage& request);


  bool   OnIncomingInstantMessage(const SIPMessageBody & iminfo);

  bool   OnReceiveOptions(const SIPRequestMessage& request,
			  SIPResponseMessage& response);

  void   OnOptionsResponse(const SIPResponseMessage& response);

  void   TraceRoute(int max, int coop);
  void initparams(MySIPCall* call);
  MySIPCall*  duplicateCall(MySIPCall* call);

  void   UseAudio(const std::string name="");
  AudioTool * GetAudioTool() {return audio; }
  bool   SetRemoteAudioAddress(const AFIAddress & addr);
  AFIAddress GetSelfAudioAddress() const;
  bool   StartAudio();
  bool   StopAudio();

  void    SetPartner(MySIPCall *call) {partner_call = call; }

  int GetCallNum() {return call_num;};

  friend int main(int argc, char *argv[]);

  bool   IsPendingIncoming() const 
  { return !isoriginator && !busy; }

  void SetBusy(bool value = true) {busy = true;}

  IptelAddress GetFloorURL() const { return floor_url; }

protected:
  /** whether the call is busy or not */
  bool busy;

  /** Call number used for user display */
  int call_num;

  // if the mode of testing should be cooperative
  int coop;

  // if options is normal options or a traceroute or ping 
  // to be used by OnOptionsResponse
  int tracert;
  
  // maximum number of hops
  int max_forward;

  // current number of hops used
  int current_max_forward;

  MySIPUserAgent & mysip;
  std::string hoplist ; 
/* qos related params */
  MySIPCall * partner_call;

  /** Audio tool for this call */
  AudioTool * audio;

  /** Floor control url */
  IptelAddress floor_url;
};

//////////////// Globals and file variables ////////////////////////////////

/**
 * The user agent can support multiple calls. The calls are stored
 * in this vector.
 */
static std::vector<MySIPCall *> calls; /* Array of calls */

/**
 * All the user global variables and parameters set using the
 * set command are stored in globals hash table.
 */
static std::map<std::string,std::string> globals;  

/**
 * Whether the software is licensed or not.
 * This may be used to limit features in unlicensed version 
 */
bool  oLicensed = false; 

/**
 * Audio tool to use.
 */
static  std::string audio_tool;
///////////////// local finle functions ////////////////////////////////////

/**
 * Display help on the commands for this user agent.
 */
static void display_help(const std::string & command);

/**
 * Access global parameter from command.
 */
static std::string global(const std::string & key);

/** 
 * Prepare the outgoing call object for initiation.
 * This function uses the global variable "globals".
 * The user contains the local user name to use, and
 * command is the command options for the command.
 */
void prepareCall(MySIPCall & mycall, CommandObject & command, 
                 const std::string user);

static void set_local_sdp(MySIPCall & mycall, std::string payloadtype = "",
                          bool ipv6 = false);

//////////////// Method defn for SIP ///////////////////////////////////////

MySIPUserAgent::MySIPUserAgent(): SIPSessionEndpoint()
{
  pthread_mutex_init(&new_call_mutex, NULL);
  pthread_cond_init(&new_call_cond, NULL);
  call_num = 0;
}

MySIPUserAgent::~MySIPUserAgent()
{
  std::cout << "~MySIPUserAgent called" << std::endl;
}

IptelCall *MySIPUserAgent::AllocateNewCallObject(const std::string callid)
{
  /* Allocate a new SIP call */
  return new MySIPCall(*this, callid);
}

void MySIPUserAgent::Wait(long timeout)
{
  struct timespec abstime;
  struct timeval now;
  gettimeofday(&now, NULL);
  abstime = timespec_delay(now, (double) timeout);

  pthread_cond_timedwait(&new_call_cond, &new_call_mutex, &abstime);
}

MySIPCall::MySIPCall(MySIPUserAgent & ep, const std::string callid)
  : SIPSession(ep, callid), mysip(ep)
{
 busy            = false;
  call_num        = 0;
  partner_call    = NULL;
  audio           = NULL;

  /* use the specified audio tool */
  UseAudio(audio_tool);

  ep.Lock();

  /* add this call to global list */
  calls.push_back(this);

  /* two digit valid call number from 1 to 99 */
  ep.call_num = (ep.call_num % 99) + 1;
  call_num = ep.call_num;
  ep.Unlock();
}

MySIPCall::~MySIPCall()
{
  MySIPUserAgent & ep = (MySIPUserAgent &) endpoint;
  std::vector<MySIPCall *>::iterator iter;

  ep.Lock();
  /* remove from list */
  for (iter = calls.begin(); iter != calls.end(); ) {
    if (*iter == this) {
      iter = calls.erase(iter);
    } 
    else {
      ++iter;
    }
  }
  ep.Unlock();

  if (partner_call != NULL) {
    partner_call = NULL; /* No not call delete */
  }

  if (audio != NULL) {
    audio->stop();
    delete audio;
  }
}

void MySIPCall::OnNewIncomingCall(IptelEndpoint & ep)
{
  if (busy) {
    Reject(486, "Busy here");
    /* How do we delete the call object here */
    return;
  }

  /*
   * Set the global variable for call.
   * Accept or reject is called in UI thread 
   */
  MySIPUserAgent & e = (MySIPUserAgent &) endpoint;

  /* prompt the user for the call */
  cout << "Incoming call [" << call_num << "] from "  
       << GetSelfAddress().AsString() << " to " 
       << GetRemoteAddress().AsString() << endl;

  if (!subject.empty()) {
    cout << "Subject is " << subject << endl;
  }
  cout << "accept/reject/noresponse ? " << flush;

  e.Lock();
  e.Signal();
  e.Unlock();
}

bool MySIPCall::OnHangup()
{
  cout << "Call [" << GetCallNum() << "] closed by remote" << endl;
  busy = false;

  // MySIPUserAgent & e = (MySIPUserAgent &) endpoint;

  StopAudio();

  SIPSession::OnHangup();
  return true;
}

void   MySIPCall::OnCallEstablished()
{
  cout << "Call [" << call_num << "] accepted by remote" << endl;
  SIPSession::OnCallEstablished();

  busy = true;

  bool start_audio = false;

  if (IsActive()) {

    SessionDescription *local = GetSelfSessionDescription();
    if (local == NULL) {
      cout << "Setting local session" << endl;
      set_local_sdp(*this);
    }

    SessionDescription *session = GetRemoteSessionDescription();
    
    if (session == NULL) {
      cout << "Remote Session is absent" << endl;
    } 
    else {
      cout << "Remote Session " << ((const SessionDescription&) *session) << endl;

#ifdef USE_GSOAP
      SessionMedia& soap = 
        remotesession->GetSession("control");
      if (soap.GetDataType() != "") {
        std::string attribute = soap.GetAttribute("a");
        if (attribute.find("floorcontrol:") == 0) {
          /* Found */
          floor_url = attribute.substr(13);
        }
      }
#endif
    
      SessionMedia & media =
        remotesession->GetSession("audio", "recv");

      AFIAddress addr = media.GetRxAddress();

      /* Check if we need to start the audio tool */
      if (partner_call == NULL) {
        /* normal call without qos measurements */
        SetRemoteAudioAddress(addr);
      }
      else {
        partner_call->SetRemoteAudioAddress(addr);
      }
      start_audio = true;
    }     
  }  

  if (start_audio == false) {
    if (partner_call == NULL) {
       StopAudio();
    }
    else {
      partner_call->StopAudio();
    }
  }
}


bool MySIPCall::OnFailureResponse(SIPResponseMessage & msg) 
{
  int status = msg.GetStatusCode();

  cout << "Response " << status << " " << msg.GetReasonPhrase() << endl;
  //cout << "Response " << status << " " << (msg.GetHeader("Warning")).AsString().c_str() << endl;

  if (status == 401 || status == 407) {
    SIPHeader h = msg.GetHeader("WWW-Authenticate");
    cout << h.AsString() << endl;
    cout << "Set your login and password and then try again." << endl;
  }
  else if (msg.Is3xxResponse()) {
    SIPContactHeader h;
    cout << h.AsString() << endl;
  }

  return SIPSession::OnFailureResponse(msg);
}

bool MySIPCall::OnCallRejected()
{
  cout << "Call [" << call_num << "] is Rejected by remote" << endl;

  busy = false;

  StopAudio();

  SIPSession::OnCallRejected();
  return true;
}

bool MySIPCall::OnRemoteSessionChanged(const SessionDescription & sdp)
{
  bool ret;

  /*
   * We always accept change in SDP
   */
  cout << "Remote SDP is changed for call [" << call_num << "], accepting" << endl;

  const AFIAddress & addr = sdp.GetConstSession().GetRxAddress();
  if (partner_call == NULL) {
    ret = SetRemoteAudioAddress(addr);
  }
  else {
    ret = partner_call->SetRemoteAudioAddress(addr);
  }
  return ret;
}

void MySIPCall::UseAudio(const std::string name)
{
  if (audio != NULL) {
    audio->stop();
    delete audio;
  }
  if (name == "") {
    /* Use the built in audio tool */
#ifdef WIN32
    audio = new WindowsAudioTool();
#else
    audio = new DeviceAudioTool();
#endif
  }
  else if (!strcasecmp(name.c_str(), "loopback")) {
    /* Loopback audio */
    audio = new LoopbackTool();
  }
  else if (!strncasecmp(name.c_str(), "file://", 7)) {
    /* Have a file audio */
    audio = new FileAudioTool(name.substr(7, name.length() - 7), "");
  }
  else {
    /* Some application */
    audio = new ExternalAudioTool(name);
  }
}

bool  MySIPCall::SetRemoteAudioAddress(const AFIAddress & addr)
{
  return (audio != NULL) ? (audio->set_destination(addr), true) : false;
}

AFIAddress MySIPCall::GetSelfAudioAddress() const
{
  return (audio != NULL) ? audio->get_receive_address() : AFIAddress();
}

bool   MySIPCall::StartAudio()
{
  return (audio != NULL) ? audio->start() : false;
}

bool   MySIPCall::StopAudio()
{
  return (audio != NULL) ? audio->stop() : false;
}

static void set_local_sdp(MySIPCall & mycall, std::string payloadtype,
                          bool ipv6)
{
  /* 
   * Allocate self session info - here, since we dont know 
   * whether the caller is ipv6 capable we will put both
   * v4 and v6
   */
  SessionDescription *session;
  session = new SessionDescription;
  SessionMedia media;
  
  media.SetDirection("recv");
  media.SetDataType("audio");
  
  if (payloadtype.empty()) {
    payloadtype = "pcmu";
  }
  
  SessionFormat f(payloadtype);
  
  AFIAddress rxaddr = mycall.GetSelfAudioAddress();
  in_port_t port = rxaddr.get_port();

  if (port == 0) {
    error("set_self_sdp", "Audio port is not allocated\n");
  }
  
  /* Add IPv6 media */
  if (ipv6) {
    media.SetAddressType("IP6");
    rxaddr = SIPLibrary::GetHostIP6Addr();
    rxaddr.set_port(port);
    media.SetRxAddress(rxaddr); 
    media.AddFormat(f); 
    session->GetSessionList().push_back(media);
  }
  
  /*. Add IPv6 media */
  media.SetAddressType("IP4");
  rxaddr = SIPLibrary::GetHostIPAddr();
  rxaddr.set_port(port);
  media.SetRxAddress(rxaddr);
  media.AddFormat(f);
  session->GetSessionList().push_back(media);

#ifdef USE_GSOAP
  std::cout << "Adding soap session" << std::endl;
  SessionMedia soap;
  SessionFormat soap_format("application/soap+xml");
  AFIAddress soap_addr(/* port = */ 9);
  soap.SetDataType("control");
  soap.SetTransportType("HTTP");
  soap.AddFormat(soap_format);
  soap.SetRxAddress(soap_addr);
  session->GetSessionList().push_back(soap);
#endif
  
  std::cout << "Session is " << *session << std::endl;

  mycall.SetSelfSessionDescription(session);
}

void prepareCall(MySIPCall & mycall, 
                 CommandObject & command, 
                 const std::string user)
{
  std::string usescheme;


  /* Set the proxy address if available in command or global. */
  std::string proxy = command["proxy"];
  if (!proxy.empty()) {
    mycall.SetProxyAddress(proxy);
    usescheme = mycall.GetProxyAddress().GetScheme();
  }
  else if (!global("proxy").empty()) {
    mycall.SetProxyAddress(global("proxy"));
    usescheme = mycall.GetProxyAddress().GetScheme();
  }

  /* Get the destination URL */
  IptelAddress dest = command.GetArg(1);

  /* Set the scheme to sip if absent in dest or proxy */
  std::string scheme = dest.GetScheme();
  if (scheme.empty()) {
    if (!usescheme.empty()) {
      dest.SetScheme(usescheme);
    } else {
      dest.SetScheme("sip");
    }
  }
  mycall.SetRemoteAddress(dest);

  /* Check if we want to use IPv6 */
  bool ipv6 = false;
  if (dest.IsIp6Address()  || 
      mycall.GetProxyAddress().IsIp6Address()) {
    ipv6 = true;
  }

  /* Formate self URL */
  IptelAddress src = command["from"];
  if (!src.IsEmpty()) {
    src.SetScheme("sip");
  }
  else {
    src = 
      (ipv6 && AFIIPv6Active()) ? SIPLibrary::GetHostName6(): SIPLibrary::GetHostName();
    src.SetScheme("sip");
    src.SetUser(user.empty() ? "sipua" : user);
  }
  mycall.SetSelfAddress(src);
  
  if (command.GetArgsCount() >= 2) {
    /* Read the request URI */
    std::string scheme;
    IptelAddress requesturi = command.GetArg(2);
    scheme = requesturi.GetScheme();
    if (scheme.empty()) {
      if (!usescheme.empty()) {
        requesturi.SetScheme(usescheme);
      } else {
        requesturi.SetScheme("sip");
      }
    }
    mycall.SetRemoteURI(requesturi);
  }

  /* Start the audio tool */
  mycall.StartAudio();

  /* Does the user want SDP in ACK instead of INVITE */
  std::string sdp = command["sdp"];
  if (sdp != "ack") {
    set_local_sdp(mycall, command["payloadtype"], ipv6);    
  }

  std::string password = global("password");
  if (!password.empty()) {
    std::string login = global("login");
    if (login.empty()) {
      login = src.GetUser();
    }
    
    if (!login.empty()) 
      mycall.SetAuthorization(login, password); 
  }

  std::string subject = global("subject");
  if (!subject.empty()) {
    mycall.SetSubject(subject);
  }

  std::string transport = command["transport"];
  if (!transport.empty()) {
    if (transport == "tcp") {
      mycall.SetPreferredTransportType(SOCK_STREAM);
    }
    else if (transport == "tls") {
      mycall.SetPreferredTransportType(SOCK_TLS);
    }
    else {
      mycall.SetPreferredTransportType(SOCK_DGRAM);
    }
  }
} /* prepareCall() */


bool MySIPCall::OnIncomingInstantMessage(const SIPMessageBody & iminfo)
{
  if (iminfo.IsContentType("text/plain")) {
    printf("IM: %s\n", iminfo.GetContent().c_str());
  } else {
    printf("IM: content-type is %s\n", iminfo.GetContentType().c_str());
  }

  return true;
}

bool MySIPCall::OnReceiveOptions(const SIPRequestMessage& request,
				 SIPResponseMessage& response)
{
  SIPMessageBody body;
  SIPHeader allow("Allow", "INVITE, ACK, CANCEL, OPTIONS, BYE, SUBSCRIBE, NOTIFY, MESSAGE, REFER");
  response.SetHeader(allow);

  SIPHeader accept_encoding("Accept-Encoding","");
  response.SetHeader(accept_encoding);
  
  SIPHeader accept_language("Accept-Language","en");
  response.SetHeader(accept_language);

  
  SIPHeader accept("Accept", "application/sdp");
  response.SetHeader(accept);
  
  std::string methods;
  int numVia =0;
  std::string acc = SIPHeader(request.GetHeader("Supported")).AsString();

	cout << acc << endl;
    
  int pos = acc.find("ping");
	//if (pos==
  
  printf("checking if it is ping or normal options %d\n",pos);

 
  if (pos!=-1)
  {
	   //cout << "OPTIONS is for Trace route , I need to do something extra\n"
	  //	<< endl;
	   
  	SIPHeader supported("Supported","application/ping");
  	response.SetHeader(supported);
           if(request.HasHeader("Via"))
	   {
		std::list<SIPHeader> via = request.GetHeaders("Via");
		for (std::list<SIPHeader>::const_iterator i = via.begin();
	  	       i != via.end(); ++i) {
		       methods += (*i).GetElement();
		       methods += "\n";
		       numVia++;
		}
	   }
 	   else
	   {
	   //	cout << "OPTIONS is for Trace route , Could not find via in options response\n"
	   //   via must be there.
	   }
	   cout << " Via's => "<< numVia << "\n " << methods << "\n"<< endl;
	   body.SetContentType("application/ping");
	   body.SetContent(methods);
	   response.SetBody(body);
   }
   


// if this is not a ping request do the following
   else {


  	SIPHeader supported("Supported","");
  	response.SetHeader(supported);



	/* Set only G.711 Mu Law audio in SDP */
	SessionDescription *session = selfsession;
	if (session == NULL) {
	    session = new SessionDescription;
	    SessionMedia media;
	    media.SetDirection("recv");
	    media.SetDataType("audio");
	    SessionFormat f("pcmu");
	    media.AddFormat(f);
	    media.SetRxAddress(AFIAddress(SIPLibrary::GetHostIPAddr()));
	    media.SetRxPort(0);
	    media.SetAddressType("IP4");
	    session->GetSessionList().push_back(media);
	  }

	  body.SetContent(session->AsString());
	  response.SetBody(body);
  }
  if (state == SIPCall::IDLE) {
    Terminate(); //terminate the dialog if not in a call.
  }

  return true;
}

void MySIPCall::OnOptionsResponse(const SIPResponseMessage& response)
{

	const char *headers[] = { "Allow", "Accept" };

	for (unsigned k=0; k<sizeof(headers)/sizeof(headers[0]); ++k) {
	    	if (response.HasHeader(headers[k])) {
 		     	std::string methods;
      			std::list<SIPHeader> allow = response.GetHeaders(headers[k]);
	      		for (std::list<SIPHeader>::const_iterator i = allow.begin();
			i != allow.end(); ++i) {
				methods += (*i).GetElement();
				methods += " ";
	      		}
    	}
  	}
  
 	if (tracert==1)
	{
	if(coop==0) // other host is no -coop
  	{
	  // if other nodes are non - cooperative.
	 	if(response.GetStatusCode()==483)
	  	{
		  //Status code received is 483, Need to check if Warning header exist
		  // Read warnign header , add it 
		  if(response.HasHeader("Warning"))
		  {

                        SIPHeader warning =    response.GetHeader("Warning");
	                         hoplist += warning.AsString();
				 hoplist+=";";
	          }
		  else
		  {
			  hoplist+="No Warning Header";
			  hoplist+=";";
		  }
		  current_max_forward++;
		  if(current_max_forward < max_forward)
		  {
                        SendOptions(current_max_forward,0);
		  }
		  else 
		  {
		  	cout << "TRACE ROUTE Non cooperative failed \n" 
			     << "The hop count  exceeded the maximum number" 
                             << " of hops specified by the user\n"
         	 	     << hoplist << endl;
                        tracert =0; // to indicate trace route is over
			max_forward =70;
		  }
	  	}
		else if(response.GetStatusCode()==200)
	  	{
		  	current_max_forward =0;
		  	cout << "TRACE ROUTE  Completed  \n" 
			      << endl;
                  	SIPHeader contact =    response.GetHeader("Contact");
			printf("%s\n",contact.AsString().c_str());
		  	std::string finallist;
	  		 char* pch = strtok ((char *)hoplist.c_str(),";");
			 while (pch != NULL)
			 {
                	    std::string hl=pch; 
	                    int pos1 = hl.find("399") + 3 ;
        	            int pos2 = hl.size();
                	    finallist=hl.substr(pos1,pos2)+"\n"+finallist;
			    //printf("%s\n", hl.substr(pos1,pos2).c_str());
	                    pch = strtok(NULL,";"); 
			 }

                	 printf("%s", finallist.c_str());
			// reinitialize variables
			 hoplist="";
		         finallist="";
			 tracert =0; // trace route is done
			 max_forward=70;
	  	}
		else 
		{
			cout << "Exiting,  Response is  " << response.GetStatusCode() 
        		<< " " << response.GetReasonPhrase() << "\n"<< endl;
			cout << " Partial Result " << "\n" << endl;
  			  char* pch = strtok ((char *)hoplist.c_str(),";");
			  while (pch != NULL)
			  {
        	            std::string hl=pch; 
                	    int pos1 = hl.find("399") + 3 ;
	                    int pos2 = hl.size();
        	  
			    printf(" => %s\n ", hl.substr(pos1,pos2).c_str());
                	    pch = strtok(NULL,";"); 
			  }
		
          	}
  	} 
  	else { // the other host is cooperative
                	 printf("Trace result completed 4\n" );
	  // get the body and parse the data 
	  if(response.GetStatusCode()==200)
	  {
                SIPHeader contact =    response.GetHeader("Contact");
		if (response.HasBody()) 
		{
		  SIPMessageBody body  = response.GetBody();
		  std::string viabody  = body.GetContent();
		  std:: string contenttype = body.GetContentType();
		  // need to parse viabody
		  cout << "TRACE ROUTE  Completed\n"
			    << endl;
                  //printf ("Splitting string \"%s\" in tokens:\n",viabody.c_str());
  		  char* pch = strtok ((char *)viabody.c_str(),";");
	          printf("%s\n",contact.AsString().c_str());
		  while (pch != NULL)
		  {
                    char * ph;
		    ph = strstr (pch,"SIP");
		    if(ph!=NULL)	
		    printf ("%s\n",ph);
		    pch = strtok (NULL, ";");
		  }

	  	  tracert =0;	
                  max_forward =70;			
		}
		else 
		{
			printf(" cooperative UA did not send via in body..retrying\n");	
			TraceRoute(max_forward,0);
		}
	  }
	  else
	  {
		  cout << "TRACE ROUTE  cooperative failed, retrying non-cooperative because\n" 
		       << response.GetStatusCode() 
		       << " " << response.GetReasonPhrase() << endl;
		
		  // start non-cooperative trace route
                   max_forward =70;
		   coop =0;
		   TraceRoute(max_forward,0);
			
	  }
	  //trecert =0;	
  	}
  	}
  

  if(tracert!=1)
  {	
  	if (state == SIPCall::IDLE) {
    		Terminate(); //terminate the dialog if not in a call.
  	}
  }


}


void MySIPCall :: TraceRoute(int maxforward,int coop_out)
{
	  tracert = 1;
	  max_forward = maxforward;
	  current_max_forward =0;
	  coop = coop_out;
          hoplist="";
	  if(coop!=0) //  the other host is cooperative
	  {
		  // the other hosts also set the content-type = application/ping
		//printf(" sending options for cooperative trace route : application/ping\n");  
		SendOptions(70,1);

	  } 
	  else  // get by normal max- forward way
	  {     // the other host is non - cooperative
     
		  //printf("sending options for non-cooperative trace route, current_max_forward %d\n",current_max_forward);
		  SendOptions(current_max_forward,0);

	  }
    

}


void MySIPCall ::initparams(MySIPCall* call)
{
     hoplist = call->hoplist;
     current_max_forward = call->current_max_forward;
     max_forward = call->max_forward;
    coop = call->coop;
    tracert = call->tracert;
}


MySIPCall*  MySIPCall::duplicateCall (MySIPCall* call)
{
        MySIPCall* mycall = new MySIPCall(call->mysip);
       mycall->initparams(call); 
	return mycall;
}


bool MySIPCall::OnSubscribe(const  SIPRequestMessage& request,
                            SIPResponseMessage& response)
{
  cout << "Incoming subscribe [" << call_num << "] from "
       << GetRemoteAddress().AsString() << ", accepting" << endl;

  return true; /* accept */
}

bool MySIPCall::OnSubscribeSuccess(const SIPResponseMessage& response)
{
  cout << "Your subscription [" << call_num << "] to "
       << GetRemoteAddress().AsString() << " succeeded." << endl;
  return false; /* do not delete the object */
}

bool MySIPCall::OnSubscribeFailed(const SIPResponseMessage& response)
{
  cout << "Your subscribe [" << call_num << "] to "
       << GetRemoteAddress().AsString() << " failed: " 
       << response.GetStatusCode() << " "
       << response.GetReasonPhrase() << endl;
  return true; /* will delete the object */
}

bool MySIPCall::OnSubscribeExpired()
{
  cout << "Your subscription [" << call_num << "] to "
       << GetRemoteAddress().AsString() << " expired." << endl;
  return true; /* will delete the object */
}

bool MySIPCall::OnNotifySuccess(const SIPResponseMessage& response)
{
  cout << "Your notify [" << call_num << "] to "
       << GetRemoteAddress().AsString() << " completed." << endl;
  return false; /* do not delete the object */
}

bool MySIPCall::OnNotifyFailed(const SIPResponseMessage& response)
{
  cout << "Your notify [" << call_num << "] to "
       << GetRemoteAddress().AsString() << " failed: " 
       << response.GetStatusCode() << " "
       << response.GetReasonPhrase() << endl;
  return true; /* will delete the object */
}

void MySIPCall::OnNotify(const SIPRequestMessage& request)
{
  SIPHeader header = request.GetHeader("Event");
  cout << "Incoming notification [" << call_num << "] from "
       << GetRemoteAddress().AsString() << " for event=\"" 
       << header.GetValue() << "\"" << endl;
  if (request.HasBody()) {
    SIPMessageBody body = request.GetBody();
    cout << "[" << body.GetContentType() << "] " << endl;
    cout << body.GetContent() << endl;
  }
}

/* Get an active call */
MySIPCall *GetActiveCall(CommandObject & command, bool pending_only = false)
{
  std::string call_num = command["call"];

  for (unsigned j = 0; j < calls.size(); ++j) {
    if (calls[j] != NULL) {
      printf("Call [%d] is present\n", calls[j]->GetCallNum());
    }
  }

  if (!call_num.empty()) {
    long num = atoi(call_num.c_str());
    for (unsigned i = 0; i < calls.size(); ++i) {
      if (calls[i]->GetCallNum() == num) {
        return calls[i];
      }
    }
  }
  else if (!pending_only) {
    if (calls.size() == 1) {
      return (*calls.begin());
    }
    else {
      return NULL;
    }
  }
  else {
    for (unsigned i = 0; i < calls.size(); ++i) {
      if (calls[i]->IsPendingIncoming()) {
        return calls[i];
      }
    }
  }

  return NULL;
}

/* print the information about the software */
static void print_software_info()
{
  cout << "SIPUA";
#ifdef VERSION
  cout << " " << VERSION;
#endif
  cout << ", (c) 2000-2003, Columbia University" << endl;
  cout << "Visit http://www.cs.columbia.edu/IRT/cinema for more info" << endl;
}

void usage(const std::string & command)
{
  const char *usage = "[options]\n"
    " -v          : Print version and exit\n"
    " -d category : verbose mode for the category. (see document)\n"
    "               all,sql,net,sdp,misc,...\n"
    " -o tracefile: put debug trace in this file.\n"
    " -a audiotool: specify a different audio tool.\n"
    " -p port     : listening SIP port, default is 5060.\n"
    " -f file     : command file to be executed.\n"
    " -n          : non interactive mode, forbackground.\n"
    " -i          : use dotted decimal for self address.\n"
    " -t          : print timing (latency) information.\n"
    " -P port     : media self port number.Default is 10000\n"
    " -H HostName : IPv6 host name, if different from normal\n"
    " -S          : Do not use SSL, to avoid annoying error messages\n"
    " -l          : Default payload type for testing. \n"
    "               0 = PCMU, 8=PCMA, etc.\n"
    ;


  cerr << "Usage: " << command << " "  << usage << endl;

  exit(1);
}

/* Main routine */
int main(int argc, char *argv[])
{
  /* the server object */
  MySIPUserAgent sip;
  /* assign the listening port */
  sip.SetListeningPort(5060);

  /* Set the server name */
  SIPLibrary::SetName("SIPUA-Columbia-University");
  /* Initialize debug options */
  SIPLibrary::ErrorOn("stderr");
  /* Override SIPLib defaults here ... */
  SIPLibrary::SetThreadPoolSize(32); /* 32 is enough here */
#ifdef USE_TLS
  SIPLibrary::StartSSL(true); /* start TLS */
#endif

  /* Print the software information */
  print_software_info();

  /* command line variables */
  std::string command_file;
  bool interactive = true;
    
  /* Parse command line options. */
  int c;
  char *cmd;
  while ((c = getopt(argc, argv, "hvd:o:a:p:f:niH:S")) != EOF) {
    switch (c) {
      case 'H' :
        SIPLibrary::SetHostName6(optarg);
        break;
      case 'v' :
#ifdef VERSION
        printf("sipua version %s\n", VERSION);
#else
        printf("sipua version unknown, not compiled with VERSION macro.\n");
#endif
        exit(0);
        break;
      case 'd':
        SIPLibrary::DebugOn();
        DebugAllowModule(optarg, 1);
        break;
      case 'o':
        SIPLibrary::DebugOn(optarg);
        break;
      case 'a':
        audio_tool = optarg;
        break;
      case 'p':
        sip.SetListeningPort(atoi(optarg));
        break;
      case 'f' :
        command_file = optarg;
        break;
      case 'n' :
        interactive = false;
        break;
      case 'i' :
        SIPLibrary::UseNumericAddress(true);
        break;
      case 'S' :
#ifdef USE_TLS
        SIPLibrary::StartSSL(false);
#else
        cerr << "Not compiled using ssl, ignoring -S option." << endl;
#endif
        break;
      case '?':
        /* no break */
      case 'h':
        usage(argv[0]);
        exit(1);
    } 
  }

  SIPLibrary::Initialize();
 
  int e =
    SIPLibrary::CheckLicense("file://sipua.license", "sipua");

  if (e == 0) {
    oLicensed = true;
  } else {
    oLicensed = false;
    cerr << "Unlicensed copy. See the web page for licensing information.\n"; 
    if (!audio_tool.empty()) {
      cerr << "Only internal device audio tool is supported in unlicensed version." << endl;
      cerr << "Ignoring -a " << audio_tool << " option" << endl;
      audio_tool = "";
    }
  }

  /* Start the listening server. */
  sip.StartServer();
  cout << "SIP " << sip << endl;

#ifdef USE_TLS
  if (SIPLibrary::StartSSL()) {
    SIPLibrary::Sleep(2);             /* the TLS lib needs sometime to start */
  }
#endif

#ifdef USE_READLINE
  /* Initialize history file */
  cout << "Using the readline library to maintain "
    "command history in file " << SIPUA_HISTORY_FILE << endl;
  if (read_history(SIPUA_HISTORY_FILE) < 0) {
    cout << "Could not open the past history file." << endl;
  }
#endif

#ifdef USE_GSOAP
  std::string logname 
    = getenv("LOGNAME") ? getenv("LOGNAME") :
      (getenv("USERNAME") ? getenv("USERNAME") : "");
  globals["soapuser"] = logname + "@" + SIPLibrary::GetHostName();
  globals["soappassword"] = "";
  globals["soapserver"] = "http://" + SIPLibrary::GetHostName() + ":8080";
#endif

  cout << "Waiting for connections..." << endl;
  cout << "Type help to get list of commands" << endl;

  bool oExitFlag = false;
  FILE *command_fp = NULL;
  while (!oExitFlag) {

    if (!command_file.empty()) {
      if (command_fp == NULL) {
        command_fp = fopen(command_file.c_str(), "r");
        if (command_fp == NULL) {
          command_file = "";
          continue;
        }
      }

      cmd = (char *)malloc(200);

      if (fgets(cmd, 199, command_fp) == NULL) {
        command_file = "";
        continue;
      }
      if (strlen(cmd) >= 1) {
        cmd[strlen(cmd)-1] = 0;
        if (cmd[0] == '#') continue; /* skip comment */
      }
    }
    else {
      if (interactive) {
#ifdef USE_READLINE
        if ((cmd = readline("sipua> ")) == NULL) {
          oExitFlag = true;
          break;
        }

        if (cmd && *cmd) {
          add_history(cmd);
        }
#else
        cout << "sipua> " << flush; /* prompt for command */
        cmd = (char *)malloc(200);
        if (fgets(cmd, 199, stdin) == NULL || feof(stdin) || ferror(stdin)) {
          oExitFlag = true;
          break;
        }
#endif
      }
      else {
        /* Do nothing in non-interactive mode. wait for ctrl-C */
        while (1) {
          SIPLibrary::Sleep(10);
        }
      }
    }

    if (cmd == NULL) 
       continue;

    CommandObject command(cmd, globals); /* Create the command object */

    if (command == "")
       continue;

    if (command == "exit" || command == "quit") {
      oExitFlag = true;
    }
    else if (command == "help") {
      std::string topic;
      if (command.GetArgsCount() > 0) {
        topic = command.GetArg(1);
      }
      display_help(topic);
    }
    else if (command ==  "?") {
      if (command.GetArgsCount() <= 0) {
        display_help(NULL);
        continue;
      }
      if (command.GetArg(1) == "call") {
        if (calls.empty()) {
          cout << "No call present" << endl;
        }
        else {
          std::vector<MySIPCall *>::iterator i;
          for (i = calls.begin(); i != calls.end(); ++i) {
            MySIPCall *call = *i;
            if (call != NULL) {
              cout << *call << endl;
            }
          }
        }
      }
      else if (command.GetArg(1) == "ep") {
        cout << sip << endl;
      }
      else if (command.GetArg(1) == "calls") {
        cout << "Numbers of calls: " << calls.size() << endl;
      }
      else if (command.GetArg(1) =="eps") {
        cout << "Number of endpoints: " << SIPLibrary::GetEndpointCount()
             << endl;
      }
      else {
        display_help(NULL);
      }
    }
    /* ---- Logging mechanism --- */
    else if (command == "log") {
      if (command.GetArgsCount() <= 0) {
        display_help("log");
        continue;
      }
      if (command.GetArg(1) == "msgflow") {
        if (command.GetArg(2) == "off") {
          SIPLibrary::MessageFlowOn(NULL);
        }
        else {
          if (command.GetArgsCount() >= 2) {
            if (command.GetArg(2) == "stdout") {
              SIPLibrary::MessageFlowOn("stdout");
            }
            else {
              SIPLibrary::MessageFlowOn("stderr");
            }
          }
          else {
            SIPLibrary::MessageFlowOn("stderr");
          }
        }
      }
      else if (command.GetArg(1) =="debug") {
        if (command.GetArg(2) == "off") {
          SIPLibrary::DebugOff();
        }
        else {
          if (command.GetArgsCount() >= 2) {
            SIPLibrary::DebugOn(command.GetArg(2));
          }
          else {
            SIPLibrary::DebugOn();
          }
        }
      }
      else if (command.GetArg(1) == "error") {
        cout << "errors log can not be modified" << endl;
      }
      else {
        display_help("log");
      }

    }
    /* ---- SIP Invite ---- */
    else if (command == "invite") {
      if (command.GetArgsCount() <= 0) {
        display_help("invite");
        continue;
      }

      /* Modified by Akira Tsukamoto 8/25/2002 */
      /* The code to setup sip session is moved to prepareCall()
       * for reusing it in "qos" command and readability */
      
      /* get login name for SIP */
      std::string logname;
      logname = getenv("LOGNAME") ? getenv("LOGNAME") : 
        (getenv("USERNAME") ? getenv("USERNAME") : "sipua");

      MySIPCall *mycall = new MySIPCall(sip);
      
      if (mycall != NULL) {
        prepareCall(*mycall, command, logname);
        mycall->Initiate();
      }
      else {
        cerr << "Memory error in allocating MySIPCall." << endl;
      }
    }
    /* "qos" command is added by Akira Tsukamoto 08/24/2002 */
    else if (command == "qos") {
      if (command.GetArgsCount() <= 0) {
        display_help("qos");
        continue;
      }

      /* Make two calls here */
      MySIPCall *mycallSend = new MySIPCall(sip);
      MySIPCall *mycallRecv = new MySIPCall(sip);

      if (mycallSend && mycallRecv) {
        mycallSend->SetPartner(mycallRecv);
        mycallRecv->SetPartner(mycallSend);

        /* First call as a Send participant */
        prepareCall(*mycallSend, command, "qos_send");
        mycallSend->Initiate();

        /* Second call as a Recv participant */
        prepareCall(*mycallRecv, command, "qos_recv");
        mycallRecv->Initiate();
      }
      else {
        if (mycallSend) delete mycallSend;
        if (mycallRecv) delete mycallRecv;
        cerr << "Memory error in allocating MySIPCall." << endl;
      }
      /* clear flags used for audio */
    }
    else if (command == "reinvite") {
      MySIPCall * call = GetActiveCall(command);
      if (call != NULL) {
        call->Reinvite();
      }
      else {
        cout << "Not an active call" << endl;
        continue;
      }
    }
    else if (command == "refer") {
      if (command.GetArgsCount() < 1) {
        display_help("refer");
        continue;
      }
      MySIPCall * call = GetActiveCall(command);
      if (call != NULL) {
        call->Refer(command.GetArg(1));
      }
      else {
        cout << "Not an active call" << endl;
        continue;
      }
    }
    else if (command == "bye") {
      MySIPCall * call = GetActiveCall(command);

      if (call != NULL) {
        call->busy = false;
        call->StopAudio();  
        call->Hangup();
        delete call;
      }
      else {
        cout << "Not an active call" << endl;
      }
    }
    else if (command == "im") {
      MySIPCall *call = GetActiveCall(command);

      if (call == NULL) {
        cerr << "No active call." << endl;
        continue;
      }
      
      call->SendInstantMessage(command.TillEnd(1));
    }
    /*-------------- Incoming responses ------------------- */
    else if (command == "waitforresponse") {
      int timeout = -1;
      if (!command["timeout"].empty()) {
        timeout = atoi(command["timeout"].c_str());
      }

      MySIPCall * call = GetActiveCall(command);

      if (call != NULL) {
        call->WaitForResponse(timeout);
      }
      else {
        cerr << "Not a valid call" << endl;
      }
    }
    else if (command == "waitforcall") {
      sip.Lock();
      if (!command["timeout"].empty()) {
        long timeout = atoi(command["timeout"].c_str());
        sip.Wait(timeout);
      } 
      else {
        sip.Wait();
      }
      sip.Unlock();
    }
    else if (command == "set" || command == "unset") {
      if (command.GetArgsCount() <= 0) {
        display_help(command.GetName());
        continue;
      }

      std::map<std::string,std::string>::iterator gi;
      gi = globals.find(command.GetArg(1));

      if (command == "set") {
        if (command.GetArgsCount() < 2) {          
          if (gi != globals.end()) {
            cout << command.GetArg(1) << " is " << (*gi).second << endl;
          }
          else {
            cout << command.GetArg(1) << " is not set" << endl;
          }
        }
        else {
          if (gi != globals.end())
            globals.erase(gi);

          globals.insert(std::make_pair(command.GetArg(1), command.TillEnd(2)));
        }
      }
      else {
        if (gi != globals.end()) 
          globals.erase(gi);
      }
    }
    else if (command == "register" || command == "unregister"
           || command == "register-get") {
      if (command.GetArgsCount() <= 0) {
        display_help("register");
        continue;
      }

      std::string usescheme;

      /* Register with SIP server */
      SIPOutgoingRegister *reg = new SIPOutgoingRegister(sip);

      /* Set the outbound proxy */
      std::string proxy = command["proxy"];
      if (!proxy.empty()) {
        reg->SetProxyAddress(proxy);
        usescheme = reg->GetProxyAddress().GetScheme();
      } 
      else if (!global("proxy").empty()) {
        reg->SetProxyAddress(global("proxy"));
        usescheme = reg->GetProxyAddress().GetScheme();
      }

      IptelAddress sipurl    = command.GetArg(1);
      if (!command["from"].empty()) {
        reg->SetSelfAddress(command["from"]);
      }
      else {
        reg->SetSelfAddress(sipurl);  /* Set the To and From */
      }
      reg->SetRemoteAddress(sipurl);

      /* Override the registrar address */
      IptelAddress registrar = sipurl;
      if (command.GetArgsCount() >= 2) {
        registrar = command.GetArg(2);
      }
      registrar.SetUser(""); /* No user name is registrar address */

      std::string scheme = registrar.GetScheme();
      if (scheme.empty()) {
        if (!usescheme.empty()) {
          registrar.SetScheme(usescheme);
        } else {
          registrar.SetScheme("sip");
        }
      }
      reg->SetRemoteURI(registrar);

      bool ipv6 = false;
      if (AFIIPv6Active() &&
          (registrar.IsIp6Address() || 
           reg->GetProxyAddress().IsIp6Address())) {
        ipv6 = true;
      }

      IptelAddress contact;

      if (!(command == "register-get")) {
        if (command["contact"].empty()) {
 
          contact = sip.BuildContactAddress(sipurl);
        } 
        else {/* user gave a contact, set it */ 
          contact = command["contact"];
        }
        
        if (command == "unregister" && contact.IsEmpty()) {
          contact = "*";
        }
      }

      if (!contact.IsEmpty()) {
        reg->SetContactAddress(contact);
      }

      /* Set the expires header */
      if (!command["expires"].empty()) {
        reg->SetRefreshRate(atoi(command["expires"].c_str()));
      }
      else if (command == "register") {
        reg->SetRefreshRate(3600); /* default is 3600 */
      }

      /* Check if there is any body */
      std::string content = command["content"];
      std::string content_type = command["content-type"];
      if (!content.empty() && content_type.empty()) {
        /* default is a perl scheme */
        content_type = "text/x-perl";
        reg->SetContentInfo(content_type, content);
      }

      if (!global("password").empty()) {
        std::string password = global("password");
        std::string login    = global("login");
        if (login.empty()) {
          login = reg->GetSelfAddress().GetUser();
        }

        if (!login.empty()) {
          reg->SetAuthorization(login, password); 
        }
      }


      std::string transport = command["transport"];
      if (!transport.empty()) {
        if (transport == "tcp") {
          reg->SetPreferredTransportType(SOCK_STREAM);
        }
        else if (transport == "tls") {
          reg->SetPreferredTransportType(SOCK_TLS);
        }
        else {
          reg->SetPreferredTransportType(SOCK_DGRAM);
        }
      }

      cout << "Registering to server as " 
           << reg->GetSelfAddress().AsString() << " at " 
           << reg->GetRemoteURI().AsString() << " with contact " 
           << contact << endl;

      reg->Register();
    }
    else if (command == "options") {
      if (command.GetArgsCount() <= 0) {
        display_help("options");
        continue;
      }
      
      MySIPCall *call = GetActiveCall(command);
     
       if (call == NULL && command.GetArgsCount() >= 1) {
        /* Create a new call object */
        call = new MySIPCall(sip);

        /* Set the outbound proxy */
        std::string proxy = command["proxy"];
        if (!proxy.empty()) {
          call->SetProxyAddress(proxy);
        } 
        else if (!global("proxy").empty()) {
          call->SetProxyAddress(global("proxy"));
        }

        /* Set self and remote addresses */
        IptelAddress sipurl    = command.GetArg(1);
        if (!command["from"].empty()) {
          call->SetSelfAddress(command["from"]);
        }
        else {
	  /* get login name for SIP */
	  IptelAddress from;
	  if (sipurl.IsIp6Address()) {
	    from = SIPLibrary::GetHostName6();
	  }
	  else {
	    from = SIPLibrary::GetHostName();
	  }
	  from.SetScheme("sip");
	  std::string logname;
	  logname = getenv("LOGNAME") ? getenv("LOGNAME") : 
	    (getenv("USERNAME") ? getenv("USERNAME") : "sipua");
	  if (!logname.empty()) {
	    from.SetUser(logname);
	  }
	  call->SetSelfAddress(from);
        }
        
        call->SetRemoteAddress(sipurl);        

        if (!global("password").empty()) {
          std::string password = global("password");
          std::string login    = global("login");
          if (login.empty()) {
            login = call->GetSelfAddress().GetUser();
          }
          
          if (!login.empty()) {
            call->SetAuthorization(login, password); 
          }
        }

        std::string transport = command["transport"];
        if (!transport.empty()) {
          if (transport == "tcp") {
            call->SetPreferredTransportType(SOCK_STREAM);
          }
          else if (transport == "tls") {
            call->SetPreferredTransportType(SOCK_TLS);
          }
          else {
            call->SetPreferredTransportType(SOCK_DGRAM);
          }
        }
      } /* if call object not found */

      if (call != NULL) {
	cout << "Sending " << command << " from " 
	     << call->GetSelfAddress().AsString() << " to "
	     << call->GetRemoteAddress().AsString()
	     << endl;
        printf("sending options message");
	call->SendOptions(70,0);
      }
      else {
	cout << "No associated call found." << endl;
      }
    }
	/*************/

    else if (command == "traceroute") {
      if (command.GetArgsCount() <= 0) {
        display_help("traceroute");
        continue;
      }
      
      MySIPCall *call = GetActiveCall(command);
     
       if (call == NULL && command.GetArgsCount() >= 1) {
        /* Create a new call object */
        call = new MySIPCall(sip);

        /* Set the outbound proxy */
        std::string proxy = command["proxy"];
        if (!proxy.empty()) {
          call->SetProxyAddress(proxy);
        } 
        else if (!global("proxy").empty()) {
          call->SetProxyAddress(global("proxy"));
        }

        /* Set self and remote addresses */
        IptelAddress sipurl    = command.GetArg(1);
        if (!command["from"].empty()) {
          call->SetSelfAddress(command["from"]);
        }
        else {
	  /* get login name for SIP */
	  IptelAddress from;
	  if (sipurl.IsIp6Address()) {
	    from = SIPLibrary::GetHostName6();
	  }
	  else {
	    from = SIPLibrary::GetHostName();
	  }
	  from.SetScheme("sip");
	  std::string logname;
	  logname = getenv("LOGNAME") ? getenv("LOGNAME") : 
	    (getenv("USERNAME") ? getenv("USERNAME") : "sipua");
	  if (!logname.empty()) {
	    from.SetUser(logname);
	  }
	  call->SetSelfAddress(from);
        }
        
        call->SetRemoteAddress(sipurl);        

        if (!global("password").empty()) {
          std::string password = global("password");
          std::string login    = global("login");
          if (login.empty()) {
            login = call->GetSelfAddress().GetUser();
          }
          
          if (!login.empty()) {
            call->SetAuthorization(login, password); 
          }
        }

        std::string transport = command["transport"];
        if (!transport.empty()) {
          if (transport == "tcp") {
            call->SetPreferredTransportType(SOCK_STREAM);
          }
          else if (transport == "tls") {
            call->SetPreferredTransportType(SOCK_TLS);
          }
          else {
            call->SetPreferredTransportType(SOCK_DGRAM);
          }
        }
      } /* if call object not found */

      if (call != NULL) {
		  cout << "Sending " << command << " from " 
	     << call->GetSelfAddress().AsString() << " to "
	     << call->GetRemoteAddress().AsString()
	     << endl;


	//cout<<" will check max-forward and cooperative\n" << endl;
	int forwardcount = 70;
		int coop =0; // coop =0 means non cooperative
		if(command["max-forward"].empty())
		{
			forwardcount =70;
			//printf("No max-forward specified.  \n");
		} 
		else 
		{
			forwardcount = atoi((command["max-forward"]).c_str());
			//printf("max-forward specified is %d.  \n",forwardcount);
		}
        if (command["cooperative"].empty())
		{
			coop=0;
			//printf("No cooperative specified .  \n");
		} else {
			coop = atoi((command["cooperative"]).c_str());
			//printf("cooperative specified is %d.  \n",coop);
		}

        	//printf("initiating sip trace route\n");
		call->TraceRoute(forwardcount,coop);
     } 

      
	  else {
		cout << "No associated call found." << endl;
   	   }
     }  /******************/

    else if (command == "authenticate") {
      MySIPCall *mycall = GetActiveCall(command, true /* incoming pending */);

      if (mycall == NULL) {
        cerr << "No incoming call." << endl;
        continue;
      }
      
      std::string method = 
        (command["method"].empty()) ? "basic" : command["method"];

      if (command["password"].empty()) {
        std::string file   = 
          (command["file"].empty()) ? "passwd" : command["file"];

        SIPLibrary::SetUserConfig(std::string("file://")+file);

        if (!mycall->Authenticate(method)) {
          cout << "Can not be authenticated, rejecting" << endl;
          mycall->Reject(); /* error code is already set */
          mycall->StopAudio();
          delete mycall;
        }
        else {
          cout << 
          "Call is authenticated\n you can accept the call (accept/reject)?" 
          << endl;
        }
      }
      else {
        std::string user = mycall->GetRemoteAddress().GetUser();
        if (!command["login"].empty()) 
          user = command["login"];
        if (!mycall->Authenticate(method, user, command["password"], command["realm"])) {
          cout << "Can not authenticate user " << user << endl;
          mycall->Reject();
          mycall->StopAudio();
          delete mycall;
        }
        else {
          cout << 
          "Call is authenticated\n you can accept the call (accept/reject)?" 
          << endl;
        }
      } 
    }
    else if (   command == "accept" 
              || command == "reject"
              || command == "noresponse" 
              || command == "redirect") {

        if (command == "noresponse") {
          /* pretend as if not received any incoming call. 
           * XXX should we delete the call object */
          continue;
        }

        MySIPCall *mycall = GetActiveCall(command, true /* incoming */);
        if (mycall == NULL) {
          cerr << "No incoming call." << endl;
          continue;
        }

        if (!command["content"].empty()) {
          std::string content, content_type = "text/plain";

          FILE *fp;
          char text[1500]; /* XXX note this limitation */
          text[0] = '\0';

#ifdef WIN32
          char * mode = "rb";
#else
          char * mode = "r";
#endif

          if ((fp = fopen(command["content"].c_str(), mode)) != NULL) {
            int length = fread(text, 1, sizeof(text) - 1, fp);
            text[length] = 0;
            if (length > 0) {
              content = text;
            }
            else {
              content = command["content"];
            }
          }
          else {
            /* File is empty use whatever is given on the command line.
            */
            content = command["content"];
          }

          if (!command["content-type"].empty()) {
            content_type = command["content-type"];
          }
          mycall->SetContentInfo(content_type, content);
        }

        if (command == "reject") {
            int status = 0;
            std::string status_str = command["status"];
            std::string reason_str;

            if (!status_str.empty()) {
              status = atoi(status_str.c_str());
            }

            if (command.GetArgsCount() >= 1) {
              reason_str = command.TillEnd(1);
            }

            cout << "Call rejecting" << endl;
            if (status < 400 || status >= 700) {
              /* Do not send back incorrect status */
              if (!reason_str.empty()) {
                status = 603; /* decline */
              }
              else {
                status = 0;   /* unknown status */
              }
            }
            if (status == 0) {
              mycall->Reject(603);
            }
            else {
              mycall->Reject(status, reason_str);
            }

            mycall->StopAudio();
            delete mycall;
            continue;
        }
        else if (command ==  "redirect") {          
          if (command.GetArgsCount() < 1) {
            display_help("redirect");
          }
          else {
            cout << "Call redirecting to " 
                 << command.GetArg(1) << endl;
            std::string reason_str = "Redirected by user";

            if (command.GetArgsCount() >= 2) {
              reason_str = command.TillEnd(2);
            }

            mycall->Redirect(IptelAddress(command.GetArg(1)), 
                            reason_str);

            mycall->StopAudio();
            delete mycall;
          }
          continue;
        }

        IptelAddress dest    = mycall->GetSelfAddress();
        IptelAddress src     = mycall->GetRemoteAddress();
        bool isv6 = false;

        if (AFIIPv6Active() &&
            (dest.IsIp6Address() || src.IsIp6Address())) {
          isv6 = true;
        }

        /* --- get remote session --- */
        SessionDescription *session;
        session = mycall->GetRemoteSessionDescription();
        if (session != NULL) {
          cout << "Remote Session is " << ((const SessionDescription&) *session) << endl;
          SessionMedia & rem_media = session->GetSession("audio","recv");

          AFIAddress rem_addr;
          rem_addr = rem_media.GetRxAddress();
          if (rem_addr.is_valid()) {
            mycall->SetRemoteAudioAddress(rem_addr);
            if (rem_addr.is_ip6_address()) {
              isv6 = true;
            }
          }
        }
        else {
          cout << "No remote session in INVITE" << endl;
        }


        /* Start the audio tool */
        mycall->StartAudio();

        /* --- Allocate self session info --- */
        session = new SessionDescription;
  
        SessionMedia media;
        media.SetDirection("recv");
        media.SetDataType("audio");

        std::string payloadtype =
          !command["payloadtype"].empty() ? command["payloadtype"] : "pcmu";
        SessionFormat f(payloadtype);
        media.AddFormat(f);
          
        AFIAddress audio_addr = mycall->GetSelfAudioAddress();
        if (AFIIPv6Active() && isv6) {
          media.SetRxAddress(AFIAddress(SIPLibrary::GetHostIP6Addr()));
          media.SetAddressType("IP6");
          media.SetRxPort(audio_addr.get_port());
          session->GetSessionList().push_back(media);
        }
        else {
         /* self IP address */
          media.SetRxAddress(AFIAddress(SIPLibrary::GetHostIPAddr()));
          media.SetRxPort(audio_addr.get_port());
          media.SetAddressType("IP4");
          session->GetSessionList().push_back(media);
        }
        
        cout << "Self Session is " << ((const SessionDescription&) *session) << endl;

        mycall->SetSelfSessionDescription(session);

        cout << "Accepting the call..." << endl;
        mycall->SetBusy(true);
        mycall->Accept();
    }
#ifdef USE_GSOAP
    else if (command == "floor") {
      FloorControl fc;
      fc.soap->userid = (char *) strdup(globals["soapuser"].c_str());
      fc.soap->passwd = (char *) strdup(globals["soappassword"].c_str());

      MySIPCall *call = GetActiveCall(command);
      if (call != NULL) {
        IptelAddress floor_url = call->GetFloorURL();

        if (!floor_url.IsEmpty()) {
          IptelAddress new_url = floor_url;
          if (floor_url.GetUser() != "") {
            new_url.SetUser("");
            strassign(&fc.soap->userid, strdup(floor_url.GetUser().c_str()));
          }
          if (floor_url.GetPassword() != "") {
            new_url.SetPassword("");
            strassign(&fc.soap->passwd, strdup(floor_url.GetUser().c_str()));
          }
          fc.endpoint = strdup(floor_url.AsString().c_str());
        }
      }
      if (fc.endpoint == NULL) {
        fc.endpoint = globals["soapserver"].c_str();
      }

      FC_HandleCommand(command, fc);
    }
#endif
    else if (command == "subscribe" || command == "unsubscribe"
           || command == "notify") {

      if (command.GetArgsCount() <= 0) {
        display_help(command.GetName());
        continue;
      }

      
      MySIPCall *call = GetActiveCall(command);

      if (call == NULL && command.GetArgsCount() >= 1) {
        /* create a new dialog. */
        call = new MySIPCall(sip);

        /* Set the outbound proxy */
        std::string proxy = command["proxy"];
        if (!proxy.empty()) {
          call->SetProxyAddress(proxy);
        } 
        else if (!global("proxy").empty()) {
          call->SetProxyAddress(global("proxy"));
        }

        /* Set self and remote addresses */
        IptelAddress sipurl    = command.GetArg(1);
        if (!command["from"].empty()) {
          call->SetSelfAddress(command["from"]);
        }
        else {
          if (command == "notify") {
            /* get login name for SIP */
            IptelAddress from;
            if (sipurl.IsIp6Address()) {
              from = SIPLibrary::GetHostName6();
            }
            else {
              from = SIPLibrary::GetHostName();
            }
            from.SetScheme("sip");
            std::string logname;
            logname = getenv("LOGNAME") ? getenv("LOGNAME") : 
              (getenv("USERNAME") ? getenv("USERNAME") : "sipua");
            if (!logname.empty()) {
              from.SetUser(logname);
            }
          }
          else {
            call->SetSelfAddress(sipurl);  /* Set the To and From same */
          }
        }
        
        call->SetRemoteAddress(sipurl);        

        if (!global("password").empty()) {
          std::string password = global("password");
          std::string login    = global("login");
          if (login.empty()) {
            login = call->GetSelfAddress().GetUser();
          }
          
          if (!login.empty()) {
            call->SetAuthorization(login, password); 
          }
        }

        std::string transport = command["transport"];
        if (!transport.empty()) {
          if (transport == "tcp") {
            call->SetPreferredTransportType(SOCK_STREAM);
          }
          else if (transport == "tls") {
            call->SetPreferredTransportType(SOCK_TLS);
          }
          else {
            call->SetPreferredTransportType(SOCK_DGRAM);
          }
        }

      } /* if not found in existing dialogs */

      /* Check if there is any body */
      SIPMessageBody body;
      body.SetContent(command["content"]);
      body.SetContentType(command["content-type"] == "" 
                         ? "text/plain" : command["content-type"]);

      std::string event = command["event"];

      cout << "Sending " << command << " from "
           << call->GetSelfAddress().AsString() << " to " 
           << call->GetRemoteAddress().AsString() 
           << endl;

      
      if (command == "subscribe") {
        call->Subscribe(event, body);
      }
      else if (command == "unsubscribe") {
        call->Subscribe(event, body, 0);
      }
      else if (command == "notify") {
        call->Notify(event, body);
      }
    }
    else if (command ==  "sleep") {
      if (command.GetArgsCount() < 1) {
        display_help("sleep");
      }
      else {
        int sec = atoi(command.GetArg(1).c_str());
        cout << 
        "Waiting for " << sec <<" seconds"<< endl;

        if (sec <= 0) {
          sec = 0;
        }

        sleep(sec);
        cout << "Wait exiting" << endl;
      }

    }
    else {
      /* Send DTMF or MESSAGE for all other things */
      std::string digits = command.GetName();
      bool isDigit = (digits.size() > 0);
      for(unsigned i = 0; i < digits.size(); ++i) {
        if (!(digits[i] >='0' && digits[i]<='9' ||
              digits[i] == '*' || digits[i] == '#')) {
          isDigit = false;
          break;
        }
      }
      
      if (isDigit) {
        cout << "sending digits to call calls: " << digits << endl;
        for (unsigned j = 0; j < calls.size(); ++j) {
          MySIPCall * mycall = calls[j];
          if (mycall) {
            AudioTool *audio = mycall->GetAudioTool();
            if ( audio ) {
              audio->rfc2833_send(digits);
            }
          }
        }
      }
      else {        
        cout << "Unknown command: " << command.GetName() 
             << ", type help to get list of commands"
             << endl;
      }
    }

    if (cmd != NULL) {
      free(cmd);
      cmd = NULL;
    }
  }

#ifdef USE_READLINE
  write_history(SIPUA_HISTORY_FILE);
#endif

  SIPLibrary::Reset();

  cout << "Exiting" << endl;
  return 0;
}

/*---------------- Utility function for displaying help -----------*/

static void display_help(const std::string & topic)
{
  static const char * mycommands =
    "General format for any command is\n"
    " <commandname>  <list-of-options> <list-of-arguments>\n"
    "Different commands are as follows:\n"
    "  invite       Make an outgoing call\n"
    "  accept       Accept an incoming call\n"
    "  reject       Reject an incoming call\n"
    "  authenticate Authenticate an incoming call\n"
    "  noresponse   Ignore an incoming call\n"
    "  redirect     Redirect an incoming call to new place\n"
    "  register     Register with SIP server\n"
    "  unregister   Unregister previous registration\n"
    "  register-get Get the registration information\n"
    "  reinvite     Send a reinvite to existing call\n"
    "  refer        Send a REFER for unattended call transfer\n"
    "  qos          To start calls for quality of service testing\n"
    "  bye          Terminate an ongoing call\n"
    "  options      Send an OPTIONS message\n"
    "  waitforresponse Wait for response from remote\n"
    "  waitforcall  Wait for an incoming call\n"
    "  subscribe    Send a SUBSCRIBE message for MWI or presence\n"
    "  unsubscribe  Send a SUBSCRIBE message with Expires: 0\n"
    "  notify       Send a NOTIFY message\n"
#ifdef USE_GSOAP
    "  floor        Floor control commands to conference server\n"
#endif
    "  sleep        Sleep for some time\n"
    "  log          Control debug trace\n"
    "  ?            Give more information about state\n"
    "  set          Set a variable or configuration item\n"
    "  unset        Unset a variable or configuration item\n"
    "  get-ogrqh    Get outgoing request header\n"
    "  get-ogrsh    Get outgoing response header\n"
    "  get-inrqh    Get incoming request header\n"
    "  get-inrsh    Get incoming response header\n"
    "  exit         Exit the user agent\n"
    "  quit         Same as exit. Exit the user agent\n"
    "Options are of the form param=value. Every command defines\n"
    " a set of supported options. Additional options are ignored.\n"
    " To get help on a particular command type\n"
    " sipua> help <command>\n"
    " For example, to get help on \"invite\", type \"help invite\"\n"
    "Visit http://www.cs.columbia.edu/IRT/cinema/doc/sipua.html for\n"
    " a complete list of commands and user manual\n";
  
  if (topic.empty() || topic == "" || topic == "help") {
    cout << mycommands << flush;
  } 
  else if (topic == "invite") {
    cout <<
    "invite [options...] sipurl [requesturi] \n"
    "    Make an outgoing call to the specified user. If requesturi is present then\n"
    "    that is used, otherwise it is set to the sipurl. sipurl becomes the To\n"
    "    header of the outgoing INVITE SIP message. Various options supported are\n"
    "    as follows: \n"
    "     *  from specifies a different from address. The default value is taken\n"
    "        from the local host name and user id. For example setting\n"
    "        from=sip:alice@wonderland will set the From header in the\n"
    "        outgoing INVITE to sip:alice@wonderland. \n"
    "     *  transport can have value as tcp or udp or tls. Default is\n"
    "        udp. \n"
    "     *  sdp can have value of ack, indicating that you want no\n"
    "        SDP in INVITE, but send in ACK.\n"
    "    invite uses the global parameters for subject, password and proxy\n"
    "    and login. \n"
    << flush;
  } else if (topic == "accept") {
    cout <<
    "accept [options...] \n"
    "    This can be used for responding to an incoming call. Following options are\n"
    "    supported. \n"
    "     *  content specifies the message body to be sent in response. \n"
    "     *  content-type specifies the Content-Type of the message body. \n"
    "     *  content-disposition specifies the Content-Disposition SIP\n"
    "        header in the response. \n"
    "     *  content-handling specifies the Content-Handling SIP header in\n"
    "        the response. \n"
    "    Note that the default handling may over-ride any of these values. In\n"
    "    particular for INVITE response the content must be SDP. \n"
    << flush;
  } else if (topic == "noresponse") {
    cout <<
    "noresponse \n"
    "    This can be used to ignore an incoming call indication. This should be used\n"
    "    for testing only. \n"
    << flush;
  } else if (topic == "reject") {
    cout <<
    "reject [options...] [reason-string] \n"
    "    This can be used to reject an incoming call. Following options are supported. \n"
    "     *  content specifies the message body to be sent in response. \n"
    "     *  content-type specifies the Content-Type of the message body. \n"
    "     *  content-disposition specifies the Content-Disposition SIP\n"
    "        header in the response. \n"
    "     *  content-handling specifies the Content-Handling SIP header in\n"
    "        the response. \n"
    "     *  status specifies the SIP status code for the returned response. \n"
    "    reason-string specifies the reason text for rejecting the call. This is\n"
    "    optional. \n"
    << flush;
  } else if (topic == "authenticate") {
    cout <<
    "authenticate [options..] \n"
    "    This allows you to authenticate an incoming INVITE. Please be advised that\n"
    "    you have to accept the call explicitly even after authentication. But if the\n"
    "    authentication fails the call is automatically rejected with 401 error code.\n"
    "    Following options are supported: \n"
    "     *  method can be either basic or digest. \n"
    "        file is the text password file path name that is used for\n"
    "        authentication. Every line in this file represents a \"login:realm:hash\".\n"
    "     *  The hash must be MD5 hash of \"login:realm:password\". See the\n"
    "        sippasswd tool or the md5string tool to generate this. The\n"
    "        incoming call request's login and password are validated against this\n"
    "        file. \n"
    << flush;
  } else if (topic == "redirect") {
    cout <<
    "redirect newurl [reason-string] \n"
    "    Use this command to redirect an incoming call to a new address.\n"
    "    reason-string is sent to remote in the 3xx response. \n"
    << flush;
  } else if (topic == "register" ||
             topic == "unregister"|| 
	           topic == "register-get") {
    cout <<
    "(register | unregister | register-get) [options...] sipurl [registrar] \n"
    "    register is used to register with a SIP registrar with the given sipurl.\n"
    "    If registrar is specified then the registration request is sent to that,\n"
    "    otherwise the registrar address is derived from sipurl. registrar is a SIP\n"
    "    URL of the form sip:hostname. On the otherhand, sipurl is of the form\n"
    "    sip:user@domain. unregister unregisters all the previous registrations\n"
    "    for this sipurl. register-get allows you to retrieve the registration\n"
    "    information from the server for this sipurl. Following options are supported\n"
    "    for these commands: \n"
    "     *  transport specifies whether to use tcp or udp for sending out\n"
    "        REGISTER message. \n"
    "     *  proxy specifies the address of the outbound proxy. This should be a\n"
    "        SIP URL of the form sip:hostname. Note that dotted IP address can also\n"
    "        be used in place of hostname. This is true for any command or option. \n"
    "     *  content allows you to set the message body of the outgoing\n"
    "        REGISTER. This is useful in uploading CGI or CPL script in\n"
    "        REGISTER message. \n"
    "     *  content-type specifies the content type. This is applicable only\n"
    "        when content is specified. Default content-type is\n"
    "        \"text/x-perl\" if content is present. \n"
    "     *  from specifies the From header address. By default the from value is\n"
    "        same as sipurl. \n"
    "     *  contact specifies the contact header. By default the contact address\n"
    "        is picked up from the local hostname and user id. You can also specify\n"
    "        * as contact for unregister command. \n"
    "     *  expires specifies the Expiry value to be used in the outgoing\n"
    "        REGISTER message. sipua takes care of refreshing the registration\n"
    "        before it expires. \n"
    << flush;
  } else if (topic == "reinvite") {
    cout <<
    "reinvite \n"
    "    reinvite sends a re-INVITE message to the remote user for the current\n"
    "    call using the existing media description. This is provided only for testing\n"
    "    purpose. \n"
    << flush;
  } else if (topic == "refer") {
    cout <<
    "refer refer_target\n"
    "    refer sends a REFER message to the specified remote user for the current call\n"
    "    This is used for unattended call transfer. The current call is terminated.\n"
    << flush;
  } else if (topic == "bye") {
    cout <<
    "bye \n"
    "    This command is used to terminate an on-going call. You also get an\n"
    "    indication when the remote party terminates the call. \n"
    << flush;
  } else if (!strncasecmp(topic.c_str(), "get-", 4)) {
    cout <<
    "(get-ogrqh | get-ogrsh | get-inrqh | get-inrsh) header \n"
    "    This is used to get the Outgoing Request Header, Outgoing Response Header,\n"
    "    Incoming Request Header and Incoming Response Header respectively. The\n"
    "    appropriate header value from the last incoming or outgoing request\n"
    "    associated with the active call is printed on the console. These commands\n"
    "    should be used only for testing purpose. \n"
    << flush;
  } else if (topic == "options") {
    cout <<
    "options url [from] \n"
    "    This allows sipua to send out SIP OPTIONS method to the specified url. from\n"
    "    specifies the From address to be used in OPTIONS. \n"
    << flush;
  } else if (topic=="traceroute") {

    cout <<
    "traceroute url [from] max-forward=[num] cooperative=[0 or any other]\n"
    "    This allows sipua to determine the path to another SIP User agent.\n"
    "    from specifies the From address to be used in OPTIONS. \n"
    "    max-forward specifies the number to be used for max-forward\n" 
    "    cooperative specifies whether  other UA is cooperative\n" 
    << flush;
  } else if (topic == "waitforresponse") {
    cout <<
    "waitforresponse\n"
    "    This command will wait for a response (success or failure) from remote\n"
    "    for an outgoing call.\n"
    << flush;
  } else if (topic == "waitforcall") {
    cout <<
    "waitforcall\n"
    "    This command will wait for a new incoming call. This is typically used\n"
    "    in automated scripts where you can accept or reject an incoming call\n"
         << flush;
  } else if (topic == "qos") {
    cout <<
    "qos [options...] sipurl [requesturi] \n"
    "    This command is similar to invite except that it makes two SIP calls to the\n"
    "    destination and invokes two external audio tools. This is used for measuring\n"
    "    the QoS of the audio call by comparing the audio packets sent in the first call\n"
    "    to that received in the second call. It is expected that the destination of the\n"
    "    call will just loopback thea audio stream, from first call to the second call.\n"
    "    For instance, if the destination is a conference server with two participants\n"
    "    invoked by this command, the audio sent by the first call is received by the\n"
    "    second. The audio tool to be used for this command is rtpqos. \n"
    << flush;
  } else if (topic == "subscribe") {
    cout <<
    "subscribe [options] sipurl\n"
    "    Send a SUBSCRIBE message for the sipurl\n"
    "    Options supported are:\n"
    "    * event  : values can be mwi or presence. Any other string is used\n"
    "               as a literal without any processing.\n"
    "    * content: message body for the outgoing SUBSCRIBE.\n"
    "    * content-type: the content type for the specified content.\n"
    "               The default content type is \"text/plain\"\n"
    << flush;
  } else if (topic == "unsubscribe") {
    cout <<
    "unsubscribe [options] sipurl\n"
    "    Send a SUBSCRIBE message with Expires: 0 header for the sipurl\n"
    "    Options supported are:\n"
    "    * event  : values can be mwi or presence. Any other string is used\n"
    "               as a literal without any processing.\n"
    "    * content: message body for the outgoing SUBSCRIBE.\n"
    "    * content-type: the content type for the specified content.\n"
    "               The default content type is \"text/plain\"\n"
    << flush;
  } else if (topic == "notify") {
    cout <<
    "notify [options] [sipurl]\n"
    "    Send a NOTIFY message in this dialog or for the sipurl\n"
    "    Options supported are:\n"
    "    * event  : values can be mwi or presence. Any other string is used\n"
    "               as a literal without any processing.\n"
    "    * content: message body for the outgoing SUBSCRIBE.\n"
    "    * content-type: the content type for the specified content.\n"
    "               The default content type is \"text/plain\"\n"
    << flush;

#ifdef USE_GSOAP
  } else if (topic == "floor") {
    cout << 
    "floor command\n"
    "    Where command is create, remove, etc.\n"
    "    Type floor without any command to get the list of commands\n"
    "    allowed.\n"
    << flush;
#endif
  } else if (topic == "sleep") {
    cout <<
    "sleep nsec \n"
    "    Sleep for nsec seconds. This is useful only when -f option is used to run the\n"
    "    commands from a file. \n"
    << flush;
  } else if (topic == "log") {
    cout <<
    "log (msgflow | debug | error) ( on [filename] | off) \n"
    "    This allows you to enable or disable debugging/trace options. msgflow option\n"
    "    displays a message flow graph for the transactions. debug is for usual verbose\n"
    "    mode trace, and error is for error display. You can either switch on or off\n"
    "    these debugging options. \n"
    << flush;
  } else if (topic == "?") {
    cout <<
    "? (call | ep | calls | eps) \n"
    "    Use this option to query the system parameters. call represents the current\n"
    "    call. ep represents the user agent server. calls and eps are count for\n"
    "    number of calls and endpoints respectively. This version supports only one\n"
    "    call and one endpoint. \n"
    << flush;
  } else if (topic == "set" || topic == "unset") {
    cout <<
    "(set | unset) (password | login | subject | proxy) [value] \n"
    "    set allows to set or retrive the value of a global parameter. If value is\n"
    "    specified then the parameter value is set to that. If it is not specified then the\n"
    "    parameter value is printed onto stdout. unset resets the value of the\n"
    "    parameter to empty. Currently following global variables are supported: \n"
    "     *  password is used for Basic or digest authentication for new call or\n"
    "        registration when the remote party requires authentication. \n"
    "     *  login is used in computing the hash for digest authentication. this is\n"
    "        useful for example, when the server asks the user to authenticate with a\n"
    "        particular login. \n"
    "     *  subject is used in the outgoing call to indicate the subject of the call.\n"
    "     *  proxy stores the SIP url of the form sip:hostname, to be used as an\n"
    "        outbound proxy. This is applicable for both outgoing call as well as\n"
    "        registration. \n"
    << flush;
  } else if (topic == "exit" || topic == "quit") {
    cout <<
    "(exit | quit) \n"
    "    This allows you to gracefully exit the application. Note that ongoing calls are\n"
    "    not explicitly terminated.     \n"
    << flush;
  } else {
    cout << "Unknown command " << topic << "\ntype help to get list of commands" << endl;
  }
          
}

/* Access the value of the global from globals map */
static std::string global(const std::string & key)
{
  std::map<std::string,std::string>::iterator gi;
  gi = globals.find(key);
  if (gi == globals.end()) {
    return "";
  }
  else {
    return (*gi).second;
  }
}
