/* sipcall.cpp - SIP Call handling.
 * 1999-2003 Columbia University. All rights reserved.
 */
#include <string>

#include "iptelcall.h"
#include "sipcallobj.h"
#include "string-util.h"
#include "siplibrary.h"
extern "C" {
#include "sysdep.h"  /* pragma for disabling warning */
#include "error.h"
}

/* construct a Call Dialog */
SIPCall::SIPCall(SIPEndpoint & ep, const std::string callid)
  : IptelCall(), SIPDialog(ep,callid)
{
  isoriginator = false;

  state = SIPCall::IDLE;

  response_initialized = false;
  response_received = false;

  empty_sdp_in_invite = true;

  subject = "Columbia SIP call";
}

/* destroy */
SIPCall::~SIPCall()
{
  /* nothing to destroy yet. 
   * Should we send BYE or CANCEL if the state is not idle.
   */
  if (state != SIPCall::IDLE && state != SIPCall::TERMINATING) {
    error("~SIPCall", "Call state is not idle or terminating.\n");
  }
}

/* 
 * Initiate an outgoing call using self and remote addresses.
 */

bool SIPCall::Initiate(const IptelAddress & to, const IptelAddress & from)
{
  SetSelfAddress(from);
  SetRemoteAddress(to);
  
  return Initiate(); /* call Initiate to initate thecall */
}

/* Initiate the call using predefined addresses. */
bool SIPCall::Initiate()
{
  switch (state) {

  case SIPCall::IDLE:
    if (remote.IsEmpty() || self.IsEmpty()) {
      error("Initiate", "Invalid self or remote address\n");
      return false;
    }

    isoriginator = true; /* outgoing request */

    /* change the state */
    state = SIPCall::OUTGOINGCALL;

    return SendInvite();
    break;

  default:
    /* ignore the message */
    return false;
    break;
  }

  return true;
}

bool SIPCall::SendInvite() {
  if (selfsession == NULL && !AllowEmptySDP()) {
    debug(DEBUG_MISC, "SIPCall::Initiate", "No SDP, using default\n");
    
    /* Set the session description to default if absent */
    selfsession =
       endpoint.GetDefaultSessionDescription();
  }
      
  SIPRequestMessage message = BuildRequest("INVITE");    
  /* set subject, body etc. */
  message.SetHeader(SIPHeader("Subject", subject));  

  if (selfsession != NULL) {
    /* Only if it is set already */
    SIPMessageBody body;
    body.SetContentType("application/sdp");
    body.SetContent(selfsession->AsString());
    message.SetBody(body);
  }
  
  SendRequest(message); /* this will create the transaction */

  return true;
}

bool SIPCall::Hangup(int code, const std::string reason)
{
  switch (state) {
  case SIPCall::OUTGOINGCALL:
    /* change the state to terminating and cancel the transaction. */
    state = SIPCall::TERMINATING;    
    SendCancel();
    break;

  case SIPCall::INCOMINGCALL:
    /* reject the call. Don't set the state here. */
    return Reject(code, reason);
    break;

  case SIPCall::INCOMINGCALL_ACK: /* fall through */
  case SIPCall::ACTIVE:
    /* change the state */
    state = SIPCall::TERMINATING;

    return SendBye();
    break;

  default:
    /* ignore */
    break;
  }
  return true;
}

bool SIPCall::SendBye()
{
  SIPRequestMessage message = BuildRequest("BYE");
  SendRequest(message); /* this will create the transaction */
  return true;
}

/**
 * Process a new request on the existing dialog.
 * subclass should derive and implement the methods.
 */
bool SIPCall::OnReceiveRequest(SIPRequestMessage & msg)
{
  debug(DEBUG_MISC, "SIPCall::OnReceiveRequest", "entered\n");

  bool ret = SIPDialog::OnReceiveRequest(msg);

  if (ret == true) {
    /* message is already handled by lower layer, so
     * probably due to an error in CSeq matching 
     * has been responded. */
    debug(DEBUG_MISC, "SIPCall::OnReceiveRequest", 
          "already handled, exiting\n");
    return ret;
  }

  switch(msg.GetMethodType()) {
  case SIPRequestMessage::INVITE:
    debug(DEBUG_MISC, "SIPCall::OnReceiveRequest", 
          "calling OnIncomingInvite\n");
    OnIncomingInvite(msg);
    ret = true;
    break;

  case SIPRequestMessage::BYE:
    debug(DEBUG_MISC, "SIPCall::OnReceiveRequest", 
          "calling OnIncomingBye\n");
    OnIncomingBye(msg);
    ret = true;
    break;

  case SIPRequestMessage::MESSAGE:
    debug(DEBUG_MISC, "SIPCall::OnReceiveRequest", 
          "calling OnIncomingInstantMessage\n");
    SendResponse(200);
    OnIncomingInstantMessage(msg.GetBody());
    ret = true;
    break;

  case SIPRequestMessage::OPTIONS: 
    {
      SIPResponseMessage response = BuildResponse(200, "OK");

      debug(DEBUG_MISC, "SIPCall::OnReceiveRequest",
	    "calling OnReceiveOptions\n");

      if (OnReceiveOptions(msg, response) == false) {
	response.SetStatusCode(501);
	response.SetReasonPhrase("Not implemented");
      }

      SendResponse(response);
      ret = true;
    }
    break;

  default:
    break;
  }

  /* return the result */
  debug(DEBUG_MISC, "SIPCall::OnReceiveRequest", 
        "exiting return=%s\n", ret ? "true" : "false");
  return ret;
}

bool SIPCall::OnIncomingInvite(SIPRequestMessage & msg)
{
  bool result = true;

  switch (state) {
  case SIPCall::IDLE:
    /* change the state */
    state = SIPCall::INCOMINGCALL;
    /* respond back with 100, to stop retransmission */
    SendResponse(100, "Trying");
    StoreRemoteSession(msg);
    debug(DEBUG_MISC, "SIPCall::OnIncomingInvite", 
          "calling OnNewIncomingCall\n");
    OnNewIncomingCall(endpoint);
    break;

  case SIPCall::ACTIVE:
  case SIPCall::INCOMINGCALL_ACK: /* So that we handle out of order 
                                   * re-INVITE */
    /* do not change the state */
    if (StoreRemoteSession(msg)) {
      /* Invoke change only if its actually changed */
      result = OnRemoteSessionChanged(*remotesession);
    }

    if (result) {
      if (selfsession == NULL) {
        debug(DEBUG_MISC, "SIPCall::Accept", "No SDP, don't send one\n");
        
        SendResponse(200, "Re-INVITE accepted");
      }
      else {
        SIPResponseMessage response = BuildResponse(200, "OK");

        if (!FillMessageSDP(response)) {
          /* Possibly mismatch in SDP */
          state = SIPCall::IDLE;

          response.SetStatusCode(488); /* SDP mismatch */
          response.SetReasonPhrase("SDP mismatch");
          response.AddHeader(SIPHeader("Accept", "application/sdp"));
          /* Don't return any error from this method */
        }          

        SendResponse(response);
      }
    }
    else {
      SendResponse(400, "Re-INVITE rejected, use original session");
    }
    break;

  case SIPCall::OUTGOINGCALL:
    error("OnIncomingInvite", "INVITE in OUTGOING CALL state\n");
    SendResponse(481, "Loop detected for INVITE");
    break;

  default:
    error("OnIncomingInvite", "INVITE in unexpected state\n");
    SendResponse(500, "Unexpected INVITE in this state");
    break;
  }

  return true;
}

bool SIPCall::OnIncomingBye(SIPRequestMessage & msg)
{
  switch (state) {
  case SIPCall::ACTIVE:
  case SIPCall::INCOMINGCALL:
  case SIPCall::INCOMINGCALL_ACK:
    /* change the state */
    state = SIPCall::TERMINATING;
    SendResponse(200, "OK"); /* accept the BYE */
    if (OnHangup()) {
      Terminate();
    }
    break;

  case SIPCall::OUTGOINGCALL:
    /* change the state */
    state = SIPCall::TERMINATING;
    SendResponse(200, "OK"); /* accept the BYE */
    if (OnCallRejected()) {
      Terminate();
    }
    break;

  case SIPCall::TERMINATING:
    /* just respond saying ok */
    SendResponse(200, "OK"); /* accept the BYE */
    break;

  case SIPCall::IDLE:
    /* Bye is no call state */
    SendResponse(482, "Call dialog does not exist");
    break;

  default:
    /* invalid state */
    error("OnIncomingBye", "BYE in unexpected state\n");
    SendResponse(500, "BYE in unexpected state");
    break;
  }

  return true;
}

/*
 * Incoming request was cancelled. don't do anything.
 * response for cancel and the original request is automatically
 * sent by the transaction later.
 */
bool SIPCall::OnCancelRequest(SIPRequestMessage & msg)
{
  bool ret = SIPDialog::OnCancelRequest(msg);

  switch(state) {
  case SIPCall::INCOMINGCALL:
  case SIPCall::INCOMINGCALL_ACK:

    state = SIPCall::TERMINATING;

    /* inform the API that call is closed */
    if (OnHangup()) {
      /* the object can be freed */
      Terminate();
    }
    break;

  default:
    error("OnCancelRequest", 
      "CANCEL received in unexpected state, ignored\n");
    break;
  }

  return ret;
}

/*
 * Process an ACK for the transaction.
 * there is no response for the ACK.
 */
bool SIPCall::OnTransactionAck(SIPRequestMessage & msg)
{
  bool ret = SIPDialog::OnTransactionAck(msg);

  switch (state) {
  case SIPCall::INCOMINGCALL_ACK:
  case SIPCall::ACTIVE:
    state = SIPCall::ACTIVE;

    /* Copy the message body */
    if (msg.HasBody() && remotesession == NULL) {
      if (StoreRemoteSession(msg)) {
        OnRemoteSessionChanged(*remotesession);
      }
    }
    /* no need to inform anyone. */
    break;

  default:
    /* All other states just ignore the message */
    error("OnTransactionAck", "ACK received in unexpected state\n");
    break;
  }

  return ret;
}

bool SIPCall::StoreRemoteSession(SIPMessage & msg)
{
  debug(DEBUG_MISC, "SIPCall::StoreRemoteSession", "entered\n");

  bool ret = false;
  SIPMessageBody body = msg.GetBody();
  std::string type = body.GetContentType();

  if (!type.empty() && !strcasecmp(type.c_str(), "application/sdp")) {
    if (remotesession != NULL) {
      delete remotesession;
    }
    remotesession = new SessionDescription;
    if (remotesession != NULL) {
      *remotesession = body.GetContent();
      /* XXX should we invoke OnRemoteSessionChanged(),
       * may be yes, but after we make sure that it actually
       * changed from the SDP version, or SDP content. */

      ret = true;
    }
    else {
      error("OnReceiveRequest", 
        "malloc error in allocating remotesession\n");
    }
  }

  debug(DEBUG_MISC, "SIPCall::StoreRemoteSession", 
        "exiting return=%s\n", ret ? "true" : "false");

  return ret;
}

/*
 * Process a new incoming response.
 * outgoing request was successful. No need to send ACK as it will 
 * be sent by SIPCall object.
 */
bool SIPCall::OnSuccessResponse(SIPResponseMessage & msg)
{
  bool ret = SIPDialog::OnSuccessResponse(msg);
  std::string method = msg.GetMethod();

  debug(DEBUG_MISC, "SIPCall::OnSuccessResponse",
        "Response %d for %s\n", msg.GetStatusCode(), method.c_str());

  /* ACK is already sent by the SIPEndpoint::OnReceiveRequest */
  if (method.empty() || method == "INVITE") {
    switch (state) {
    case SIPCall::OUTGOINGCALL:
      state = SIPCall::ACTIVE;
      /* Copy the message body */
      StoreRemoteSession(msg);
      OnCallEstablished();
      break;
      
    case SIPCall::ACTIVE:
      StoreRemoteSession(msg);
      break;
      
    default:
      /* probably a reinvite's success */
      debug(DEBUG_MISC, "OnSuccessResponse", "In unexpected state\n");
      break;
    }
  }
  else if (method == "OPTIONS") {
    OnOptionsResponse(msg);
  }

  return ret;
}

/*
 * Process a provisional response.
 */
bool SIPCall::OnProvisionalResponse(SIPResponseMessage & msg)
{
  bool ret = SIPDialog::OnProvisionalResponse(msg);
  std::string method = msg.GetMethod();

  debug(DEBUG_MISC, "SIPCall::OnProvisionalResponse",
        "Response %d for %s\n", msg.GetStatusCode(), method.c_str());

  if (method.empty() || method == "INVITE") {
    switch (state) {
    case SIPCall::OUTGOINGCALL:
      /* we do not support early media */
      StoreRemoteSession(msg);
      OnAlerting();
      break;
      
    default:
      /* just ignore the message */
      break;
    }
  }

  return ret;
}

/**
 * Process a failure response.
 */
bool SIPCall::OnFailureResponse(SIPResponseMessage & msg)
{
  bool ret = SIPDialog::OnFailureResponse(msg);
  std::string method = msg.GetMethod();

  debug(DEBUG_MISC, "SIPCall::OnFailureResponse",
        "Response %d for %s\n", msg.GetStatusCode(), method.c_str());

  if (method.empty() || method == "INVITE") {
    switch (state) {
    case SIPCall::OUTGOINGCALL:
      state = SIPCall::TERMINATING;
      if (OnCallRejected()) {
        Terminate();
      }
      break;
      
    default:
      /* reinvite failed, state does not change. */
      error("OnFailureResponse", "In unexpected state, ignoring\n");
      break;
    }
  }
  else if (method == "OPTIONS") {
    OnOptionsResponse(msg);
  }

  return ret;
}

/**
 * Process a timeout
 * Since the processing is same as SIPDialog's callback, we
 * do not define it here
bool SIPCall::OnTimerEvent(void *token)
{
  bool ret = SIPDialog::OnTimerEvent(token);
  return ret;
}
 */


bool SIPCall::Alert()
{
  switch(state) {
  case SIPCall::INCOMINGCALL:
    /* send the 180 response */
    SendResponse(180, "Alerting");
    break;

  default:
    /* ignore in other states */
    return false;
  }
  return true;
}

bool SIPCall::Alert(int status, const std::string reason)
{
  if (status < 100 || status >= 200) {
    /* not a ringing message. */
    return false;
  }

  switch(state) {
  case SIPCall::INCOMINGCALL:
    /* send the status response */
    if (selfsession == NULL) {
      SendResponse(status, reason);
    }
    else {
      SIPResponseMessage message =
        BuildResponse(status);
      SIPMessageBody body;

      body.SetContentType("application/sdp");
      body.SetContent(selfsession->AsString());

      message.SetBody(body);
      SendResponse(message);
    }
    break;

  default:
    /* ignore in other states */
    return false;
  }
  return true;
}

/* reinvite */
bool SIPCall::Reinvite()
{
  switch(state) {
  case SIPCall::IDLE:
    /* Initiate the call */
    return Initiate();

  case SIPCall::ACTIVE:
  case SIPCall::INCOMINGCALL_ACK:
    return SendInvite();
    break;

  default:
    return false;
  }

  return true;
}

bool SIPCall::Accept()
{
  SIPResponseMessage message;

  switch (state) {
  case SIPCall::INCOMINGCALL:
    state = SIPCall::INCOMINGCALL_ACK; /* wait for ACK */
    
    if (selfsession == NULL) {
      debug(DEBUG_MISC, "SIPCall::Accept", "No SDP, using default\n");
      
      /* Set the session description to default if absent */
      selfsession = 
         endpoint.GetDefaultSessionDescription();
    }

    /* build response */
    message = BuildResponse(200, "OK");
    
    if (FillMessageSDP(message)) {
      SendResponse(message);
    }
    else {
      state = SIPCall::IDLE;

      message.SetStatusCode(488); /* SDP mismatch */
      message.SetReasonPhrase("SDP mismatch");
      message.AddHeader(SIPHeader("Accept", "application/sdp"));
      
      SendResponse(message);

      if (OnHangup()) {
        Terminate();
      }
      return false;
    }

    break;

  default:
    return false; /* all other states. */
  }

  return true;
}

bool SIPCall::Reject()
{
  switch (state) {
  case SIPCall::INCOMINGCALL:
    state = SIPCall::IDLE; /* wait for ACK */
    SendResponse(0);
    break;

  default: /* all other states */
    return false;
  }
  return true;
}

bool SIPCall::Reject(int status, const std::string & reason)
{
  if (status < 300) {
    /* incorrect status code */
    return false;
  }

  switch (state) {
  case SIPCall::INCOMINGCALL:
    state = SIPCall::IDLE; /* wait for ACK */
    SendResponse(status, reason);
    break;

  default: /* all other states */
    return false;
  }
  return true;
}

bool SIPCall::Redirect(const IptelAddress newaddress,
                      const std::string & reason)
{
  SIPContactHeader contact;
  SIPResponseMessage message;

  /* Set the contact header, send 302 response. */
  switch (state) {
  case SIPCall::INCOMINGCALL:
    state = SIPCall::IDLE; /* wait for ACK */
    message = BuildResponse(302, reason);

    contact.SetAddress(newaddress);
    
    message.SetHeader(contact);

    SendResponse(message);
    break;

  default: /* all other states */
    return false;
  }
  return true;
}



/** 
 * Set the content type and the content body if different
 * from default (application/sdp).
 */
void SIPCall::SetContentInfo(const std::string & contentType,
                            const std::string & contentBody)
{
  self_content.SetContent(contentBody);
  self_content.SetContentType(contentType);
}
  
/**
 * Method to send Instant Message 
 */
bool SIPCall::SendInstantMessage(std::string text, std::string type)
{
  SIPMessageBody body;

  body.SetContentType(type);
  body.SetContent(text);

  return SendInstantMessage(body);
}

bool SIPCall::SendInstantMessage(const SIPMessageBody& body)
{
  SIPRequestMessage msg = BuildRequest("MESSAGE");
  
  msg.SetBody(body);

  SendRequest(msg);
  return true;
}

/** 
 * Callback for incoming MESSAGE message.
 */
bool SIPCall::OnIncomingInstantMessage(const SIPMessageBody & iminfo)
{
  return false;
}

/**
 * Method to send OPTIONS
 */
bool SIPCall::SendOptions(unsigned int max_forwards,int tracert)
{
  	SIPRequestMessage msg = BuildRequest("OPTIONS");
  	
  	if (max_forwards >= 0) {
    	/* Add 1 to Max-Forwards before sending since libsip automatically
       	reduces by one. */
    		SIPHeader mf("Max-Forwards", int2str(max_forwards+1));
    		msg.SetHeader(mf);
  	}
  // supported header used to indiacte that request is of type application ping. 
  if(tracert==1)
  {
  	SIPHeader supported("Supported", "application/ping");
  	msg.SetHeader(supported);
  }

  SendRequest(msg);
  return true;
}

bool SIPCall::OnReceiveOptions(const SIPRequestMessage& request,
			       SIPResponseMessage& response)
{
  return false;
}

void SIPCall::OnOptionsResponse(const SIPResponseMessage& response)
{
  return;
}

/**
 * This callback is called on error.
 */
void   SIPCall::OnError(const std::string & reason)
{
  if (!reason.empty()) {
    error("SIPCall", "%s\n", reason.c_str());
  }
}

/**
 * Waiting for a final response from remote for outgoing call.
 */
int    SIPCall::WaitForResponse(int timeout)
{
  if (response_initialized == false) {
    pthread_cond_init(&response_cond, NULL);
    pthread_mutex_init(&response_mutex, NULL);
    response_initialized = true;
  }
  pthread_mutex_lock(&response_mutex);
  while (!response_received) {
    /* XXX use the timeout for waiting */
    pthread_cond_wait(&response_cond, &response_mutex);
  }
  pthread_mutex_unlock(&response_mutex);
  return 0;
}

/**
 * Block till call is terminated.
 */
int    SIPCall::WaitForTermination(int timeout)
{
  /*
   * XXX use logic similar to WaitForResponse to implement this. 
   */
  return 0;
}

/**
 * Callback on new incoming call.
 */
void   SIPCall::OnNewIncomingCall(IptelEndpoint & ep)
{
  /* not an originator of the call */
  isoriginator = false;
  /* default implementation just rejects the call */
  Reject(500, "The method is not implemented in the sub-class");
  return; /* do nothing else */
}

/**
 * Indication that remote phone is ringing.
 */
void   SIPCall::OnAlerting()
{
  return; /* do nothing here */
}

/**
 * Callback on successful call establishment.
 */
void   SIPCall::OnCallEstablished()
{
  SignalResponseReceived();
}


/**
 * Callback on call rejected by remote.
 */
bool  SIPCall::OnCallRejected()
{
  SignalResponseReceived();
  return true;
}

/**
 * Callback on call hung up.
 */
bool  SIPCall::OnHangup()
{
  SignalResponseReceived();
  return true;
}


/**
 * Signal to all the waiting threads if needed, that response is
 * received for this call.
 */
void SIPCall::SignalResponseReceived()
{
  if (response_initialized) {
    pthread_mutex_lock(&response_mutex);
    response_received = true;
    pthread_cond_broadcast(&response_cond);
    pthread_mutex_unlock(&response_mutex);
  }
}

/**
 * Callback when the remote session is changed.
 */
bool SIPCall::OnRemoteSessionChanged(const SessionDescription & sdp)
{
  return true;
}

/**
 * Fill the SDP using the selfsession information, into the message
 * body. Return true on success, or false on SDP mismatch.
 *
 * @param message is the SIPMessage in which the body is filled using
 * SDP of selfsession.
 *
 * @return true on success, and false on SDP mismatch. 
 */
bool SIPCall::FillMessageSDP(SIPMessage & message)
{
  std::string reason;
  SessionDescription *newsession = NULL;
  SIPMessageBody body;

  /* Get the subset of session description */
  if (remotesession == NULL) {
    reason = "No session present";
    
    body.SetContentType("application/sdp");
    body.SetContent(selfsession->AsString());
    message.SetBody(body);
  }
  else {
    newsession = selfsession->MatchWithSource(*remotesession, reason);
    
    if (newsession == NULL) {
      /* there was a mismatch */
      if (!reason.empty()) {
        message.AddHeader(SIPHeader("Warning", reason));
      }      
      return false;
    }
    else {
      body.SetContentType("application/sdp");
      body.SetContent(newsession->AsString());
      message.SetBody(body);
      delete newsession;
    }
  }  

  return true;
}
