/* sipcallobj.h - SIP API specification
 * (c) 1999-2003 Columbia University. All rights reserved.
 */
#ifndef _SIP_CALLOBJ_H_
#define _SIP_CALLOBJ_H_

#include <string>
#include <list>

#include "iptelcall.h"         /* IPtelEndPoint definitions */
#include "sipmessage.h"
#include "siptransaction.h"
#include "sipdialog.h"

//Forward declaration.
class SIPEndpoint;
class SessionDescription;

/**
 * This represents a SIP call between self and remote.
 * Various methods and attributes allow changing call state and
 * retrieving call information.  A call is associated with an
 * #IptelEndpoint#. It is required that the endpoint be of type
 * #SIPEndpoint# or a derived class.
 *
 * @see SIPDialog, IptelCall, SIPEndpoint, SessionDescription
 */
class SIPCall : public IptelCall, public SIPDialog
{
 public:
  /**@name Construct and destroy */
  //@{
  /** 
   * Create a new call for the endpoint.
   * This attaches the SIP call with the endpoint
   * specified. Some of the attributes can be
   * borrowed (for example, self name) from the
   * associated endpoint.
   */
  SIPCall(SIPEndpoint & ep, const std::string callid = "");
  
  /** Destructor to destroy the object */
  virtual ~SIPCall();
  //@}

  /**@name Call state control */
  //@{
  /**
   * Place an outgoing call with already set
   * self and remote addresses.
   *
   * @return true on successful initiation and false on failure.
   * Not that even if the method returns true the actual call may fail,
   * and OnCallRejected or OnCallEstablished will be called as 
   * appropriate.
   */
  virtual bool Initiate();
  
  /**
   * Place an outgoing call. 
   * To initiate a new outgoing call with the given
   * remote and self addresses. 
   *
   * @param to Specifies the self (from) address. This is of the
   * form of "From:" SIP header.
   * @param from specifies the remote (to) address.
   * This is of the form of "To:" SIP header.
   *
   * @return true on successful initiation and false on failure. 
   */
  virtual bool  Initiate(const IptelAddress & to,
                         const IptelAddress & from);
  
  /**
   * Terminate a call.
   *
   * @param reason is the status code.
   * @param strReason is the reason string. 
   *
   * @return true on success and false on failure.
   */
  virtual bool  Hangup(int reason = 0, /* 0 for normal termination */
                       const std::string strReason = "Normal Termination");
  
  /**
   * Indicate that the local phone is ringing to remote.
   * This sends 180 Ringing response to remote for
   * an incoming call.
   *
   * @return true on success and false on failue.
   */
  virtual bool  Alert();
  
  /**
   * Indicate that the local phone is ringing to remote
   * using a specific status code. If the session description is
   * set then it uses the SDP in the message body. This
   * is different from Alert() with no arguments in which
   * case the SDP is never sent. This method should be used
   * for early media or when alerting status other than 180
   * needs to be sent.
   *
   * @return true on success and false on failue.
   */
  virtual bool  Alert(int status, /* 183 for example */
                      const std::string reason = "Ringing");
  
  /**
   * Accept an incoming call.
   *
   * @return true on success and false on failure.
   */
  virtual bool  Accept();
  
  /**
   * Reject an incoming call. Use the pre-built status code
   * or "603" not already set.
   *
   * @return true on success and false on failue.
   */
  virtual bool  Reject();
  
  /**
   * Reject an incoming call with appropriate reason
   * 
   * @param status is the status code and .
   * @param reason is the reason string to be  sent
   *        in the SIP response to the caller.
   *
   * @return true on success and false on failue.
   */
  virtual bool  Reject(int status, const std::string & reason="");
  
  /**
   * Redirect an incoming call with appropriate reason.
   * 
   * @param newaddress is the "Contact:" address of the new
   * destination where you want the call to be redirected.
   *
   * @param reason specifies the reason for redirection.
   * If this is NULL then default is "Redirect".
   *
   * @return true on success and false on failure.
   */
  virtual bool Redirect(const IptelAddress newaddress,
                        const std::string & reason="");
  
  /**
   * Sends INVITE message to the same destination.
   * Typically used when session description is changed
   * during a call, for example in H323v1.0->GW->SIP.
   * Note that  unlike forking of INVITE in earlier request,
   * the re-INVITE is sent to only one destination.
   * This can be called either by the originator or the
   * receiver of the original INVITE request.
   *
   * @return bool on successful initiation and false on failure.
   */
  virtual bool Reinvite();
  
  /**
   * Callback on new incoming call.
   * The library indicates an incoming call to the
   * application by calling this method with the
   * associated IptelEndpoint. The IptelEndpoint
   * muct have been created as a server listening
   * for incoming calls. Application should Accept or
   * Reject the incoming call either in the implementation
   * of this call back function or in a separate thread.
   * Default behaviour accepts any incoming call.
   * 
   * @param ep is the associated endpoint object
   * with the call.
   *
   * @see #Accept#, #Reject#.
   */
  virtual void   OnNewIncomingCall(IptelEndpoint & ep);
  
  /**
   * Indication that remote phone is ringing.
   * When a 180 response is received the Library 
   * calls this callback to indicate remote ringing to the
   * application. Application can generate local
   * ring tones as appropriate.
   * Default behaviour does nothing.
   */
  virtual void   OnAlerting();
  
  /**
   * Waiting for a final response from remote for outgoing call.
   * Application can call this method to block till 
   * a final response is received for an outgoing call.
   *
   * @param timeout in seconds can be specified. Default is -1 for
   * INFINITE timeout.
   */
  virtual int    WaitForResponse(int timeout = -1);
  
  /**
   * Callback on successful call establishment.
   * The Library calls this method to indicate successful call
   * establishment. Default behaviour does nothing.
   */
  virtual void   OnCallEstablished();
  
  /**
   * Callback on call rejected by remote.
   * The Library calls this method to indicate that the
   * remote has rejected your call.
   * @return true if you want the library to free the call object.
   * otherwise return non-zero if your application is responsible
   * for freeing the call object.
   */
  virtual bool  OnCallRejected();
  
  /**
   * Callback on call hung up.
   * The library calls this method to indicate that 
   * an ongoing call was hungup by remote.
   *
   * @return true if you want the library to free the call object.
   * otherwise the application is responsible for
   * freeing the call object.
   */
  virtual bool  OnHangup();
  
  /**
   * Block till call is terminated.
   * Application can call this method to block till
   * the call is terminated by remote. Note that
   * if you are using this method then OnCallRejected and
   * OnHangup should return non-zero to avoid bus-error.
   *
   * @param timeout in seconds can be specified.
   * Default is -1 for INFINITE timeout.
   *
   * @return 0 on success and -1 on failure.
   */
  virtual int    WaitForTermination(int timeout = -1);
  
  /**
   * This callback is called on error.
   * Application can act on the error, for example,
   * by printing the error message and terminating the call.
   * 
   * @param errorno indicates the error which occurred.
   * @see IptelCall::strerror.
   */
  virtual void   OnError(const std::string & reason);
  
  /**
   * Is it an outgoing or incoming call. This is defined by
   * the direction of the first transaction in this dialog.
   * @return true if this is the originator of the call.
   * false otherwise.
   */
  virtual bool   IsOriginator() const { return isoriginator;}
  
  //@}

  /**@name Instant messaging. */
  //@{
  /**
   * Method to send Instant Message 
   * @return true: Succeed, false: Fail
   */
  virtual bool SendInstantMessage(std::string text, std::string type="text/plain");

  /**
   * Method to send Instant Message.
   * @return true: succeed, false: fail
   */
  virtual bool SendInstantMessage(const SIPMessageBody& body);

  /** 
   * Callback for incoming MESSAGE message.
   * This is a callback invoked by the stack to inform the
   * application of a new incoming MESSAGE.
   */
  virtual bool OnIncomingInstantMessage(const SIPMessageBody & iminfo);
  //@}
  
  /**@name OPTIONS related methods */
  //@{
  /**
   * Send a SIP OPTIONS message in the dialog.
   */
  bool SendOptions(unsigned int max_forwards = 75, int tracert = 0);

  /**
   * Callback when an incoming options is received. The application
   * should set the response message. The return value indicates
   * whether the processing is successful (i.e., response is valid) or
   * not. If the processing is not successful (i.e., it returns false)
   * then the library sends a "501 Not implemented" response
   * automatically. The default implementation just returns false.  If
   * the application returns true, then the library sets the response
   * message body (if not already set) to the SDP of the call object
   * (if available).
   */
  virtual bool OnReceiveOptions(const SIPRequestMessage& request,
				SIPResponseMessage& response);

  /**
   * Callback when the outbound OPTIONS has received a response.  The
   * library internally calls the callback even if there is a timeout.
   * The default implementation does nothing.
   */
  virtual void OnOptionsResponse(const SIPResponseMessage& response);

  //@}

  /**@name Misc methods */
  //@{
  /** 
   * Set the content type and the content body if different
   * from default (application/sdp).
   * @param contentType application/sdp, application/html.
   * @param contentBody appropriate body, NULL terminated string.
   */
  virtual void SetContentInfo(const std::string & contentType,
                              const std::string & contentBody);
    
  /** Set the subject header */
  void SetSubject(const std::string & s) {subject = s;}
  
  /** Retrieve the subject header, return NULL if none. */
  std::string GetSubject() const {return subject;}
  
  /**
   * Callback when the remote session is changed.
   * Note that the session description should not be freed by the 
   * application.
   * @return the application should return true if accepts the
   * session change and false if not.
   */
  virtual bool OnRemoteSessionChanged(const SessionDescription & sdp);

  /**
   * Whether to allow the empty SDP in the outgoing INVITE
   * message. Otherwise the library puts a default SDP in the outgoing
   * message. It doesn't affect the incoming INVITE message as
   * incoming INVITE without SDP is always allowed. By default the
   * SIPCall allows empty SDP in outgoing INVITE.
   *
   * @param allow includes whether empty SDP is allowed in outgoing INVITE.
   */
  void AllowEmptySDP(bool allow) {empty_sdp_in_invite = allow; }

  /**
   * Whether no SDP is allowed in INVITE.
   *
   * @return true if no SDP is allowed in INVITE, else false.
   */
  bool AllowEmptySDP() const { return empty_sdp_in_invite; }

  /**
   * Is the call active. 
   *
   * @return true if the call is in active state, else return false.
   */
  virtual bool IsActive() const { return state == ACTIVE;}

  /**
   * Is the call already established. This is either the ACTIVE or
   * INCOMINGCALL-ACK state where the application is waiting for
   * remote to send ACK but we assume that the call is established
   * already.
   *
   * @return true if teh call is in active or incomingcall-ack state,
   * else return false. 
   */
  virtual bool IsEstablished() const { return state == ACTIVE || state == INCOMINGCALL_ACK; }

  //@}

  /**@name Overridden from SIPDialog 
     @see SIPDialog 
  */
  //@{
  /**
   * Process a new request on the existing transaction.
   * The default behavior invokes OnNewIncomingCall
   * for a new call, or OnRemoteSessionChanged for
   * re-INVITE if the session changes, or OnHangup if this is
   * a BYE request. For other requests just invoke
   * SIPDialog's callback.
   *
   * @param msg is the request message that is received.
   *
   * @return true if success, and false on failure. Application
   * should never return false, unless it wants to cleanup the
   * dialog abnormally after the callback.
   */
  virtual bool OnReceiveRequest(SIPRequestMessage & msg);

  /**
   * Process a new incoming cancel request for a transaction. 
   * If this was for a INVITE and the call is in incoming
   * state then OnHangup() is invoked otherwise just
   * invoke the SIPDialog's callback.
   *
   * @param msg is the CANCEL request received for this dialog.
   *
   * @return true if success, and false on failure. Application
   * should never return false, unless it wants to cleanup the
   * dialog abnormally after the callback.
   */
  virtual bool OnCancelRequest(SIPRequestMessage & msg);

  /**
   * Process an ACK for the transaction. Change the state
   * of the call and do invoke the SIPDialog's callback.
   *
   * @param msg is the ACK message received for this dialog.
   *
   * @return true if success, and false on failure. Application
   * should never return false, unless it wants to cleanup the
   * dialog abnormally after the callback.
   */
  virtual bool OnTransactionAck(SIPRequestMessage & msg);

  /**
   * Process a new incoming response.
   * Depending on the state call OnCallEstablished. 
   * At the end call the SIPDialog's callback.
   *
   * @param msg is the response message received for this dialog.
   *
   * @return true if success, and false on failure. Application
   * should never return false, unless it wants to cleanup the
   * dialog abnormally after the callback.
   */
  virtual bool OnSuccessResponse(SIPResponseMessage & msg);

  /**
   * Process a provisional response.
   * Depending on the state call OnAlerting().
   * At the end call the SIPDialog's callback.
   *
   * @param msg is the response message received for this dialog.
   *
   * @return true if success, and false on failure. Application
   * should never return false, unless it wants to cleanup the
   * dialog abnormally after the callback.
   */
  virtual bool OnProvisionalResponse(SIPResponseMessage & msg);

  /**
   * Process a failure response.
   * Depending on the state call OnHangup().
   * At the end call the SIPDialog's callback.
   *
   * @param msg is the response message received for this dialog.
   *
   * @return true if success, and false on failure. Application
   * should never return false, unless it wants to cleanup the
   * dialog abnormally after the callback.
   */
  virtual bool OnFailureResponse(SIPResponseMessage & msg);

  //@}

 protected:
  /** format and send an INVITE message */
  virtual bool SendInvite();

  /** format and send a BYE */
  virtual bool SendBye();

  /** incoming INVITE */
  virtual bool OnIncomingInvite(SIPRequestMessage & msg);

  /** incoming BYE */
  virtual bool OnIncomingBye(SIPRequestMessage & msg);

  /** store the remotesession if present in the message */
  bool StoreRemoteSession(SIPMessage & msg);

  /**
   * Fill the SDP using the selfsession information, into the message
   * body. Return true on success, or false on SDP mismatch. In case
   * of SDP mismatch it adds a Warning header in the message.
   *
   * @param message is the SIPMessage in which the body is filled using
   * SDP of selfsession.
   *
   * @return true on success, and false on SDP mismatch. 
   */
  bool FillMessageSDP(SIPMessage & message);
    
  /** false for incoming call. true for outgoing */
  bool isoriginator;
  
  /** subject of the call */
  std::string subject;

  /** Message body if given */
  SIPMessageBody self_content;

  /** Allow empty SDP in INVITE, default is true. */
  bool empty_sdp_in_invite;

 /** Call states */
 enum SIPCallState {
   IDLE = 0, 
   OUTGOINGCALL, 
   INCOMINGCALL, 
   INCOMINGCALL_ACK, 
   ACTIVE,
   TERMINATING
 } state;

 private:
   /**
    * Signal to all the waiting threads if needed, that response is
    * received for this call.
    */
   void SignalResponseReceived();

   /* for receiving responses */
   bool response_initialized;
   bool response_received;
   pthread_cond_t  response_cond;
   pthread_mutex_t response_mutex;

};

#endif /* _SIP_CALLOBJ_H_ */

/************************** END OF FILE ********************************
*/
